"""Tests for DOS."""

import numpy as np

from phonopy import Phonopy
from phonopy.phonon.dos import get_pdos_indices

tp_str = """0.000000 100.000000 200.000000 300.000000 400.000000
500.000000 600.000000 700.000000 800.000000 900.000000
4.856373 3.916036 -0.276031 -6.809284 -14.961974
-24.342086 -34.708562 -45.898943 -57.796507 -70.313405
0.000000 26.328820 55.258118 74.269718 88.156943
99.054231 108.009551 115.606163 122.200204 128.024541
0.000000 36.207859 45.673598 47.838871 48.634251
49.009334 49.214960 49.339592 49.420745 49.476502"""

tdos_str = """-0.750122 0.000000
0.249878 0.000041
1.249878 0.010398
2.249878 1.363019
3.249878 1.837252
4.249878 0.991171
5.249878 0.342658
6.249878 0.106961
7.249878 0.032966"""

tdos_thm_str = """-0.750122 0.000000
0.249878 0.004508
1.249878 0.090716
2.249878 0.549237
3.249878 1.093288
4.249878 1.349766
5.249878 0.660090
6.249878 0.461219
7.249878 0.000000"""

pdos_str = """-0.750122 0.000000 0.000000
0.249878 0.000016 0.000025
1.249878 0.004469 0.005929
2.249878 0.620921 0.742098
3.249878 0.597587 1.239666
4.249878 0.501077 0.490094
5.249878 0.142162 0.200496
6.249878 0.070026 0.036935
7.249878 0.020225 0.012741"""

pdos_thm_str = """-0.750122 0.000000 0.000000
0.249878 0.001820 0.002688
1.249878 0.037419 0.053297
2.249878 0.228129 0.321109
3.249878 0.403637 0.689651
4.249878 0.756532 0.593234
5.249878 0.454573 0.205518
6.249878 0.322527 0.138692
7.249878 0.000000 0.000000"""


def testTotalDOS(ph_nacl_nofcsym: Phonopy):
    """Test of total DOS with smearing method."""
    phonon = ph_nacl_nofcsym
    phonon.run_mesh([5, 5, 5])
    phonon.run_total_dos(freq_pitch=1, use_tetrahedron_method=False)
    dos = phonon.total_dos.dos
    freqs = phonon.total_dos.frequency_points
    data_ref = np.reshape([float(x) for x in tdos_str.split()], (-1, 2))
    np.testing.assert_allclose(data_ref, np.c_[freqs, dos], atol=1e-5)
    # for f, d in zip(freqs, dos):
    #     print("%f %f" % (f, d))


def testTotalDOSTetrahedron(ph_nacl_nofcsym: Phonopy):
    """Test of total DOS with tetrahedron method."""
    phonon = ph_nacl_nofcsym
    phonon.run_mesh([5, 5, 5])
    phonon.run_total_dos(freq_pitch=1, use_tetrahedron_method=True)
    dos = phonon.total_dos.dos
    freqs = phonon.total_dos.frequency_points
    data_ref = np.reshape([float(x) for x in tdos_thm_str.split()], (-1, 2))
    np.testing.assert_allclose(data_ref, np.c_[freqs, dos], atol=1e-5)
    # for f, d in zip(freqs, dos):
    #     print("%f %f" % (f, d))


def testProjectedlDOS(ph_nacl_nofcsym: Phonopy):
    """Test projected DOS with smearing method."""
    phonon = ph_nacl_nofcsym
    phonon.run_mesh([5, 5, 5], is_mesh_symmetry=False, with_eigenvectors=True)
    phonon.run_projected_dos(freq_pitch=1, use_tetrahedron_method=False)
    pdos = phonon.projected_dos.projected_dos
    freqs = phonon.projected_dos.frequency_points
    data_ref = np.reshape([float(x) for x in pdos_str.split()], (-1, 3)).T
    np.testing.assert_allclose(data_ref, np.vstack([freqs, pdos]), atol=1e-5)
    # for f, d in zip(freqs, pdos.T):
    #     print(("%f" + " %f" * len(d)) % ((f, ) + tuple(d)))


def testPartialDOSTetrahedron(ph_nacl_nofcsym: Phonopy):
    """Test projected DOS with tetrahedron method."""
    phonon = ph_nacl_nofcsym
    phonon.run_mesh([5, 5, 5], is_mesh_symmetry=False, with_eigenvectors=True)
    phonon.run_projected_dos(freq_pitch=1, use_tetrahedron_method=True)
    pdos = phonon.projected_dos.projected_dos
    freqs = phonon.projected_dos.frequency_points
    data_ref = np.reshape([float(x) for x in pdos_thm_str.split()], (-1, 3)).T
    np.testing.assert_allclose(data_ref, np.vstack([freqs, pdos]), atol=1e-5)
    # for f, d in zip(freqs, pdos.T):
    #     print(("%f" + " %f" * len(d)) % ((f, ) + tuple(d)))


def test_get_pdos_indices(ph_tio2: Phonopy):
    """Test get_pdos_indices by TiO2."""
    indices = get_pdos_indices(ph_tio2.primitive_symmetry)
    np.testing.assert_array_equal(indices[0], [0, 1, 2, 3])
    np.testing.assert_array_equal(indices[1], [4, 5])



"""Tests for routines in force_constants.py."""

import numpy as np
import pytest

from phonopy import Phonopy
from phonopy.harmonic.force_constants import (
    cutoff_force_constants,
    rearrange_force_constants_array,
)
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import get_primitive, isclose


def test_fc(ph_nacl: Phonopy):
    """Test of force constants calculation with fcsym by NaCl."""
    fc_1_10_ref = [
        -0.037549,
        0.000000,
        0.000000,
        0.000000,
        0.002415,
        -0.001746,
        0.000000,
        -0.001746,
        0.002415,
    ]
    fc_1_10 = ph_nacl.force_constants[1, 10]
    # print("".join(["%f, " % v for v in fc_1_10.ravel()]))
    np.testing.assert_allclose(fc_1_10.ravel(), fc_1_10_ref, atol=1e-5)


def test_fc_nofcsym(ph_nacl_nofcsym: Phonopy):
    """Test of force constants calculation without fcsym by NaCl."""
    fc_1_10_nofcsym_ref = [
        -0.005051,
        0.000000,
        0.000000,
        0.000000,
        0.094457,
        0.000000,
        0.000000,
        0.000000,
        -0.020424,
    ]

    fc_1_10 = ph_nacl_nofcsym.force_constants[1, 10]
    # print("".join(["%f, " % v for v in fc_1_10.ravel()]))
    np.testing.assert_allclose(fc_1_10.ravel(), fc_1_10_nofcsym_ref, atol=1e-5)


def test_fc_compact_fcsym(ph_nacl_compact_fcsym: Phonopy):
    """Test of force constants calculation in compact format with fcsym by NaCl."""
    fc_1_10_compact_fcsym_ref = [
        -0.004481,
        0.000000,
        0.000000,
        0.000000,
        0.095230,
        0.000000,
        0.000000,
        0.000000,
        -0.019893,
    ]

    fc_1_10 = ph_nacl_compact_fcsym.force_constants[1, 10]
    # print("".join(["%f, " % v for v in fc_1_10.ravel()]))
    np.testing.assert_allclose(fc_1_10.ravel(), fc_1_10_compact_fcsym_ref, atol=1e-5)


def test_fc_compact_symfc(ph_nacl_rd_symfc: Phonopy):
    """Test of force constants calculation in compact format using symfc by NaCl."""
    fc_1_10_compact_symfc_ref = [
        -6.777705e-03,
        0,
        0,
        0,
        9.591639e-02,
        0,
        0,
        0,
        -2.087195e-02,
    ]
    fc_1_10 = ph_nacl_rd_symfc.force_constants[1, 10]
    # print("".join(["%f, " % v for v in fc_1_10.ravel()]))
    np.testing.assert_allclose(fc_1_10.ravel(), fc_1_10_compact_symfc_ref, atol=1e-5)


@pytest.mark.parametrize("is_compact", [False, True])
def test_fc_cutoff_radius(ph_nacl: Phonopy, ph_nacl_compact_fcsym: Phonopy, is_compact):
    """Test of cutoff radius of force constants calculation by NaCl."""
    if is_compact:
        ph = ph_nacl_compact_fcsym
    else:
        ph = ph_nacl

    # Need restore fc because fc are overwritten.
    fc_orig = ph.force_constants.copy()
    ph.set_force_constants_zero_with_radius(4.0)
    changed = np.abs(fc_orig - ph.force_constants) > 1e-8
    ph.force_constants = fc_orig

    if is_compact:
        assert np.sum(changed) == 534
    else:
        assert np.sum(changed) == 17088


@pytest.mark.parametrize(
    "is_compact,store_dense_svecs",
    [(False, False), (False, True), (True, False), (True, True)],
)
def test_fc_cutoff_radius_svecs(
    ph_nacl: Phonopy, ph_nacl_compact_fcsym: Phonopy, is_compact, store_dense_svecs
):
    """Test of cutoff radius with dense-svecs format by NaCl."""
    if is_compact:
        ph = ph_nacl_compact_fcsym
    else:
        ph = ph_nacl

    fc = ph.force_constants.copy()
    primitive_matrix = np.dot(np.linalg.inv(ph.supercell_matrix), ph.primitive_matrix)
    primitive = get_primitive(
        ph.supercell, primitive_matrix, store_dense_svecs=store_dense_svecs
    )

    cutoff_force_constants(fc, ph.supercell, primitive, 4.0)
    changed = np.abs(ph.force_constants - fc) > 1e-8

    if is_compact:
        assert np.sum(changed) == 534
    else:
        assert np.sum(changed) == 17088


def test_rearrange_force_constants_array(
    nacl_unitcell_order1: PhonopyAtoms,
    nacl_unitcell_order2: PhonopyAtoms,
    ph_nacl: Phonopy,
):
    """Test of rearrange_force_constants_array."""
    ph1 = ph_nacl
    assert isclose(nacl_unitcell_order1, ph1.unitcell)
    fc = ph1.force_constants
    assert fc.shape[0] == fc.shape[1]
    ph2 = Phonopy(
        nacl_unitcell_order2,
        supercell_matrix=[2, 2, 2],
        primitive_matrix="F",
    )
    re_fc, indices = rearrange_force_constants_array(fc, ph1.supercell, ph2.supercell)
    ph1.run_qpoints([[0.5, 0.5, 0.5]])
    freq1 = ph1.get_qpoints_dict()["frequencies"]
    ph2.force_constants = re_fc
    ph2.run_qpoints([[0.5, 0.5, 0.5]])
    freq2 = ph2.get_qpoints_dict()["frequencies"]
    np.testing.assert_allclose(fc[indices[3], indices[3]], re_fc[3, 3])
    np.testing.assert_allclose(freq1, freq2)



"""SSCHA calculation."""

# Copyright (C) 2024 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import copy
from typing import Optional

import numpy as np

from phonopy import Phonopy
from phonopy.interface.mlp import PhonopyMLP
from phonopy.physical_units import get_physical_units


class MLPSSCHA:
    """Iterative approach SSCHA using MLP."""

    def __init__(
        self,
        ph: Phonopy,
        mlp: PhonopyMLP,
        temperature: Optional[float] = None,
        number_of_snapshots: Optional[int] = None,
        max_iterations: Optional[int] = None,
        distance: Optional[float] = None,
        fc_calculator: Optional[str] = None,
        log_level: int = 0,
    ):
        """Init method.

        ph : Phonopy
            Phonopy instance.
        mlp : PhonopyMLP
            PhonopyMLP instance.
        temperature : float, optional
            Temperature in K, by default 300.0.
        number_of_snapshots : int, optional
            Number of snapshots, by default 2000.
        max_iterations : int, optional
            Maximum number of iterations, by default 10.
        distance : float, optional
            Distance of displacements, by default is None, which gives 0.01.
        fc_calculator : str, optional
            Force constants calculator. The default is None, which means "symfc".
        log_level : int, optional
            Log level, by default 0.

        """
        if mlp is None:
            raise ValueError("MLP is not provided.")

        if temperature is None:
            self._temperature = 300.0
        else:
            self._temperature = temperature
        if number_of_snapshots is None:
            self._number_of_snapshots = 1000
        else:
            self._number_of_snapshots = number_of_snapshots
        if max_iterations is None:
            self._max_iterations = 10
        else:
            self._max_iterations = max_iterations
        self._max_iterations = max_iterations
        if distance is None:
            self._distance = 0.01
        else:
            self._distance = distance
        if fc_calculator is None:
            self._fc_calculator = "symfc"
        else:
            self._fc_calculator = fc_calculator
        self._log_level = log_level

        self._ph = ph.copy()
        self._ph.mlp = PhonopyMLP(mlp=mlp.mlp)
        self._ph.nac_params = copy.deepcopy(ph.nac_params)

        # Calculate supercell energy without displacements
        self._ph.generate_displacements(distance=0, number_of_snapshots=1)
        self._ph.evaluate_mlp()
        self._supercell_energy = float(self._ph.supercell_energies[0])
        self._ph.dataset = None

        if ph.force_constants is None:
            self._iter_counter = 0
        else:
            if log_level:
                print("Use provided force constants.")
                print("")
            self._ph.force_constants = ph.force_constants
            self._iter_counter = 1

    @property
    def phonopy(self) -> Phonopy:
        """Return Phonopy instance."""
        return self._ph

    @property
    def free_energy(self) -> float:
        """Return free energy in eV."""
        return self._free_energy

    @property
    def force_constants(self) -> np.ndarray:
        """Return force constants."""
        return self._ph.force_constants

    @property
    def harmonic_potential_energy(self) -> float:
        """Return supercell energies."""
        d = self._ph.displacements
        pe = np.einsum("ijkl,mik,mjl", self.force_constants, d, d) / len(d) / 2
        return pe

    @property
    def potential_energy(self) -> float:
        """Return potential energy."""
        return np.average(self._ph.supercell_energies - self._supercell_energy)

    def calculate_free_energy(self, mesh: float = 100.0) -> float:
        """Calculate SSCHA free energy."""
        self._ph.run_mesh(mesh=mesh)
        self._ph.run_thermal_properties(temperatures=[self._temperature])
        hfe = (
            self._ph.get_thermal_properties_dict()["free_energy"][0]
            / get_physical_units().EvTokJmol
        )
        n_cell = len(self._ph.supercell) / len(self._ph.primitive)
        pe = self.potential_energy / n_cell
        hpe = self.harmonic_potential_energy / n_cell
        self._free_energy = hfe + pe - hpe

    def run(self) -> "MLPSSCHA":
        """Run through all iterations."""
        for _ in self:
            if self._log_level:
                print("")
        return self

    def __iter__(self) -> "MLPSSCHA":
        """Iterate over force constants calculations."""
        return self

    def __next__(self) -> int:
        """Calculate next force constants."""
        if self._iter_counter == self._max_iterations + 1:
            self._iter_counter = 0
            raise StopIteration
        self._run()
        self._iter_counter += 1
        return self._iter_counter - 1

    def _run(self) -> Phonopy:
        if self._log_level and self._iter_counter == 0:
            print(
                f"[ SSCHA initialization (rd={self._distance}, "
                f"n_supercells={self._number_of_snapshots}) ]",
                flush=True,
            )
        if self._log_level and self._iter_counter > 0:
            print(f"[ SSCHA iteration {self._iter_counter} / {self._max_iterations} ]")
            print(
                f"Generate {self._number_of_snapshots} supercells with displacements "
                f"at {self._temperature} K",
                flush=True,
            )

        if self._iter_counter == 0:
            self._ph.generate_displacements(
                distance=self._distance, number_of_snapshots=self._number_of_snapshots
            )
        else:
            self._ph.generate_displacements(
                number_of_snapshots=self._number_of_snapshots,
                temperature=self._temperature,
            )
            hist, bin_edges = np.histogram(
                np.linalg.norm(self._ph.displacements, axis=2), bins=10
            )

            if self._log_level:
                size = np.prod(self._ph.displacements.shape[0:2])
                for i, h in enumerate(hist):
                    length = round(h / size * 100)
                    print(
                        f"  [{bin_edges[i]:4.3f}, {bin_edges[i + 1]:4.3f}] "
                        + "*" * length
                    )

        if self._log_level:
            print("Evaluate MLP to obtain forces using pypolymlp", flush=True)

        self._ph.evaluate_mlp()

        if self._log_level:
            print("Calculate force constants using symfc", flush=True)
        self._ph.produce_force_constants(
            fc_calculator="symfc",
            fc_calculator_log_level=self._log_level if self._log_level > 1 else 0,
            calculate_full_force_constants=True,
            show_drift=False,
        )



"""Tests for routines in grid_points.py."""

import os

import numpy as np
import pytest

from phonopy import Phonopy
from phonopy.interface.phonopy_yaml import read_cell_yaml
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.grid_points import (
    GeneralizedRegularGridPoints,
    GridPoints,
    length2mesh,
)

current_dir = os.path.dirname(os.path.abspath(__file__))

ga234 = [
    [0, 0, 0],
    [1, 0, 0],
    [0, 1, 0],
    [1, 1, 0],
    [0, -1, 0],
    [-1, -1, 0],
    [0, 0, 1],
    [1, 0, 1],
    [0, 1, 1],
    [1, 1, 1],
    [0, -1, 1],
    [1, 2, 1],
    [0, 0, 2],
    [1, 0, 2],
    [0, 1, 2],
    [1, 1, 2],
    [0, -1, -2],
    [-1, -1, -2],
    [0, 0, -1],
    [-1, 0, -1],
    [0, 1, -1],
    [-1, -2, -1],
    [0, -1, -1],
    [-1, -1, -1],
]


def test_GridPoints():
    """Test of GridPoints."""
    gp = GridPoints([2, 3, 4], [[-1, 1, 1], [1, -1, 1], [1, 1, -1]])

    assert gp.ir_grid_points.dtype == np.dtype("int64")
    assert gp.weights.dtype == np.dtype("int64")
    assert gp.grid_mapping_table.dtype == np.dtype("int64")
    assert gp.grid_address.dtype == np.dtype("intc")
    assert gp.mesh_numbers.dtype == np.dtype("intc")
    assert gp.reciprocal_lattice.dtype == np.dtype("double")
    assert gp.qpoints.dtype == np.dtype("double")

    np.testing.assert_array_equal(gp.grid_address, ga234)


def test_GridPoints_NaCl_with_rotations(ph_nacl: Phonopy):
    """Test of GridPoints with rotations from NaCl."""
    rec_lat = np.linalg.inv(ph_nacl.primitive.cell)
    rotations = ph_nacl.primitive_symmetry.pointgroup_operations
    gp = GridPoints([4, 4, 4], rec_lat, rotations=rotations)
    np.testing.assert_array_equal(gp.ir_grid_points, [0, 1, 2, 5, 6, 7, 10, 27])
    np.testing.assert_array_equal(gp.weights, [1, 8, 4, 6, 24, 12, 3, 6])
    np.testing.assert_array_equal(
        gp.grid_mapping_table,
        [
            0,
            1,
            2,
            1,
            1,
            5,
            6,
            7,
            2,
            6,
            10,
            6,
            1,
            7,
            6,
            5,
            1,
            5,
            6,
            7,
            5,
            1,
            7,
            6,
            6,
            7,
            6,
            27,
            7,
            6,
            27,
            6,
            2,
            6,
            10,
            6,
            6,
            7,
            6,
            27,
            10,
            6,
            2,
            6,
            6,
            27,
            6,
            7,
            1,
            7,
            6,
            5,
            7,
            6,
            27,
            6,
            6,
            27,
            6,
            7,
            5,
            6,
            7,
            1,
        ],
    )
    np.testing.assert_allclose(
        gp.qpoints,
        [
            [0.0, 0.0, 0.0],
            [0.25, 0.0, 0.0],
            [0.5, 0.0, 0.0],
            [0.25, 0.25, 0.0],
            [0.5, 0.25, 0.0],
            [-0.25, 0.25, 0.0],
            [0.5, 0.5, 0.0],
            [-0.25, 0.5, 0.25],
        ],
        atol=1e-8,
    )


@pytest.mark.parametrize("fit_in_BZ", [True, False])
def test_GridPoints_NaCl_with_rotations_fit_BZ(ph_nacl: Phonopy, fit_in_BZ):
    """Test of GridPoints with rotations from NaCl and fit_in_BZ."""
    rec_lat = np.linalg.inv(ph_nacl.primitive.cell)
    rotations = ph_nacl.primitive_symmetry.pointgroup_operations
    mesh = [5, 5, 5]
    gpts = GridPoints(mesh, rec_lat, rotations=rotations, fit_in_BZ=fit_in_BZ)
    if fit_in_BZ:
        np.testing.assert_allclose(
            gpts.qpoints,
            [
                [0.0, 0.0, 0.0],
                [0.2, 0.0, 0.0],
                [0.4, 0.0, 0.0],
                [0.2, 0.2, 0.0],
                [0.4, 0.2, 0.0],
                [-0.4, 0.2, 0.0],
                [-0.2, 0.2, 0.0],
                [0.4, 0.4, 0.0],
                [-0.4, -0.6, 0.0],
                [0.6, 0.4, 0.2],
            ],
            atol=1e-8,
        )
    else:
        np.testing.assert_allclose(
            gpts.qpoints,
            [
                [0.0, 0.0, 0.0],
                [0.2, 0.0, 0.0],
                [0.4, 0.0, 0.0],
                [0.2, 0.2, 0.0],
                [0.4, 0.2, 0.0],
                [-0.4, 0.2, 0.0],
                [-0.2, 0.2, 0.0],
                [0.4, 0.4, 0.0],
                [-0.4, 0.4, 0.0],
                [-0.4, 0.4, 0.2],
            ],
            atol=1e-8,
        )


def test_GridPoints_SnO2_with_rotations(ph_sno2: Phonopy):
    """Test of GridPoints with rotations from SnO2."""
    rec_lat = np.linalg.inv(ph_sno2.primitive.cell)
    rotations = ph_sno2.primitive_symmetry.pointgroup_operations
    gp = GridPoints([4, 4, 4], rec_lat, rotations=rotations)
    np.testing.assert_array_equal(
        gp.ir_grid_points,
        [0, 1, 2, 5, 6, 10, 16, 17, 18, 21, 22, 26, 32, 33, 34, 37, 38, 42],
    )
    np.testing.assert_array_equal(
        gp.weights, [1, 4, 2, 4, 4, 1, 2, 8, 4, 8, 8, 2, 1, 4, 2, 4, 4, 1]
    )
    np.testing.assert_array_equal(
        gp.grid_mapping_table,
        [
            0,
            1,
            2,
            1,
            1,
            5,
            6,
            5,
            2,
            6,
            10,
            6,
            1,
            5,
            6,
            5,
            16,
            17,
            18,
            17,
            17,
            21,
            22,
            21,
            18,
            22,
            26,
            22,
            17,
            21,
            22,
            21,
            32,
            33,
            34,
            33,
            33,
            37,
            38,
            37,
            34,
            38,
            42,
            38,
            33,
            37,
            38,
            37,
            16,
            17,
            18,
            17,
            17,
            21,
            22,
            21,
            18,
            22,
            26,
            22,
            17,
            21,
            22,
            21,
        ],
    )
    np.testing.assert_allclose(
        gp.qpoints,
        [
            [0.0, 0.0, 0.0],
            [0.25, 0.0, 0.0],
            [0.5, 0.0, 0.0],
            [0.25, 0.25, 0.0],
            [0.5, 0.25, 0.0],
            [0.5, 0.5, 0.0],
            [0.0, 0.0, 0.25],
            [0.25, 0.0, 0.25],
            [0.5, 0.0, 0.25],
            [0.25, 0.25, 0.25],
            [0.5, 0.25, 0.25],
            [0.5, 0.5, 0.25],
            [0.0, 0.0, 0.5],
            [0.25, 0.0, 0.5],
            [0.5, 0.0, 0.5],
            [0.25, 0.25, 0.5],
            [0.5, 0.25, 0.5],
            [0.5, 0.5, 0.5],
        ],
        atol=1e-8,
    )


def test_GridPoints_SnO2_with_rotations_MP(ph_sno2: Phonopy):
    """Test of GridPoints with non-gamma-centre mesh and rotations from SnO2."""
    rec_lat = np.linalg.inv(ph_sno2.primitive.cell)
    rotations = ph_sno2.primitive_symmetry.pointgroup_operations
    gp = GridPoints([4, 4, 4], rec_lat, rotations=rotations, is_gamma_center=False)
    np.testing.assert_array_equal(gp.ir_grid_points, [0, 1, 5, 16, 17, 21])
    np.testing.assert_array_equal(gp.weights, [8, 16, 8, 8, 16, 8])
    np.testing.assert_array_equal(
        gp.grid_mapping_table,
        [
            0,
            1,
            1,
            0,
            1,
            5,
            5,
            1,
            1,
            5,
            5,
            1,
            0,
            1,
            1,
            0,
            16,
            17,
            17,
            16,
            17,
            21,
            21,
            17,
            17,
            21,
            21,
            17,
            16,
            17,
            17,
            16,
            16,
            17,
            17,
            16,
            17,
            21,
            21,
            17,
            17,
            21,
            21,
            17,
            16,
            17,
            17,
            16,
            0,
            1,
            1,
            0,
            1,
            5,
            5,
            1,
            1,
            5,
            5,
            1,
            0,
            1,
            1,
            0,
        ],
    )
    np.testing.assert_allclose(
        gp.qpoints,
        [
            [0.125, 0.125, 0.125],
            [0.375, 0.125, 0.125],
            [0.375, 0.375, 0.125],
            [0.125, 0.125, 0.375],
            [0.375, 0.125, 0.375],
            [0.375, 0.375, 0.375],
        ],
        atol=1e-8,
    )


@pytest.mark.parametrize("suggest", [True, False])
def test_SNF_from_GeneralizedRegularGridPoints(ph_tio2: Phonopy, suggest):
    """Test for grid rotation matrix and SNF by TiO2."""
    grgp = GeneralizedRegularGridPoints(
        ph_tio2.unitcell, 60, suggest=suggest, x_fastest=False
    )

    if suggest:
        np.testing.assert_array_equal(
            grgp.grid_matrix, [[0, 16, 16], [16, 0, 16], [6, 6, 0]]
        )
        np.testing.assert_array_equal(grgp.snf.P, [[0, -1, 3], [1, 0, 0], [-3, 3, -8]])
        np.testing.assert_array_equal(grgp.snf.D, [[2, 0, 0], [0, 16, 0], [0, 0, 96]])
        np.testing.assert_array_equal(grgp.snf.Q, [[1, 8, 17], [0, 0, -1], [0, 1, 1]])
        np.testing.assert_allclose(
            grgp.transformation_matrix,
            [[-0.5, 0.5, 0.5], [0.5, -0.5, 0.5], [0.5, 0.5, -0.5]],
        )
        assert (grgp.grid_address[253] == [0, 2, 61]).all()
        np.testing.assert_allclose(
            grgp.qpoints[253], [-0.19791667, 0.36458333, -0.23958333]
        )
        np.testing.assert_array_equal(
            grgp.reciprocal_operations[1], [[9, 10, 3], [8, 9, 3], [-48, -54, -17]]
        )
    else:
        np.testing.assert_array_equal(
            grgp.grid_matrix, [[16, 0, 0], [0, 16, 0], [0, 0, 6]]
        )
        np.testing.assert_array_equal(grgp.snf.P, [[1, 0, -3], [0, -1, 0], [-3, 0, 8]])
        np.testing.assert_array_equal(grgp.snf.D, [[2, 0, 0], [0, 16, 0], [0, 0, 48]])
        np.testing.assert_array_equal(
            grgp.snf.Q, [[-1, 0, -9], [0, -1, 0], [-1, 0, -8]]
        )
        np.testing.assert_allclose(grgp.transformation_matrix, np.eye(3))
        assert (grgp.grid_address[253] == [0, 5, 13]).all()
        np.testing.assert_allclose(grgp.qpoints[253], [-0.4375, -0.3125, -0.16666667])
        np.testing.assert_array_equal(
            grgp.reciprocal_operations[1], [[9, -1, 3], [-8, 0, -3], [-24, 3, -8]]
        )


@pytest.mark.parametrize("suggest", [True, False])
def test_GeneralizedRegularGridPoints_rotations_tio2(ph_tio2, suggest):
    """Test for GeneralizedRegularGridPoints by TiO2."""
    matches = _get_matches(ph_tio2, suggest, True)
    if suggest:
        matches_ref = [
            0,
            29,
            58,
            67,
            96,
            5,
            34,
            43,
            72,
            81,
            10,
            39,
            48,
            77,
            86,
            15,
            24,
            53,
            62,
            91,
            38,
            47,
            76,
            85,
            14,
            23,
            52,
            61,
            90,
            19,
            28,
            57,
            66,
            95,
            4,
            33,
            42,
            71,
            80,
            9,
            56,
            65,
            94,
            3,
            32,
            41,
            70,
            99,
            8,
            37,
            46,
            75,
            84,
            13,
            22,
            51,
            60,
            89,
            18,
            27,
            74,
            83,
            12,
            21,
            50,
            79,
            88,
            17,
            26,
            55,
            64,
            93,
            2,
            31,
            40,
            69,
            98,
            7,
            36,
            45,
            92,
            1,
            30,
            59,
            68,
            97,
            6,
            35,
            44,
            73,
            82,
            11,
            20,
            49,
            78,
            87,
            16,
            25,
            54,
            63,
        ]
    else:
        matches_ref = [
            0,
            25,
            40,
            15,
            30,
            5,
            20,
            45,
            10,
            35,
            2,
            27,
            42,
            17,
            32,
            7,
            22,
            47,
            12,
            37,
            4,
            29,
            44,
            19,
            34,
            9,
            24,
            49,
            14,
            39,
            6,
            21,
            46,
            11,
            36,
            1,
            26,
            41,
            16,
            31,
            8,
            23,
            48,
            13,
            38,
            3,
            28,
            43,
            18,
            33,
        ]
    np.testing.assert_array_equal(matches, matches_ref)


@pytest.mark.parametrize("suggest", [True, False])
@pytest.mark.parametrize("is_time_reversal", [True, False])
def test_GeneralizedRegularGridPoints_rotations_zr3n4(
    ph_zr3n4, suggest, is_time_reversal
):
    """Test for GeneralizedRegularGridPoints by non-centrosymmetric Zr3N4."""
    matches = _get_matches(ph_zr3n4, suggest, is_time_reversal)
    if suggest:
        matches_ref = [
            0,
            17,
            10,
            3,
            14,
            7,
            29,
            22,
            33,
            26,
            19,
            30,
            52,
            45,
            38,
            49,
            42,
            41,
            36,
            53,
            46,
            39,
            50,
            43,
            11,
            4,
            15,
            8,
            1,
            12,
            34,
            27,
            20,
            31,
            24,
            23,
            18,
            35,
            28,
            21,
            32,
            25,
            47,
            40,
            51,
            44,
            37,
            48,
            16,
            9,
            2,
            13,
            6,
            5,
        ]
    else:
        matches_ref = [
            0,
            2,
            1,
            9,
            11,
            10,
            18,
            20,
            19,
            6,
            8,
            7,
            15,
            17,
            16,
            24,
            26,
            25,
            3,
            5,
            4,
            12,
            14,
            13,
            21,
            23,
            22,
        ]
    np.testing.assert_array_equal(matches, matches_ref)


def _get_matches(ph, suggest, is_time_reversal):
    grgp = GeneralizedRegularGridPoints(
        ph.unitcell,
        20,
        suggest=suggest,
        is_time_reversal=is_time_reversal,
        x_fastest=False,
    )
    rot_address = np.dot(grgp.grid_address, grgp.reciprocal_operations[1])
    diag_n = np.diagonal(grgp.snf.D)
    rot_address %= diag_n
    matches = []
    for adrs in rot_address:
        d = np.abs(grgp.grid_address - adrs).sum(axis=1)
        match = np.where(d == 0)[0]
        assert len(match) == 1
        matches.append(match[0])
    return matches


def test_watch_GeneralizedRegularGridPoints(ph_tio2: Phonopy, helper_methods):
    """Test for q-points positions obtained from GeneralizedRegularGridPoints."""
    grgp = GeneralizedRegularGridPoints(ph_tio2.unitcell, 10, x_fastest=False)
    tmat = grgp.transformation_matrix
    # direct basis vectors in row vectors
    plat = np.dot(tmat.T, ph_tio2.unitcell.cell)
    # reciprocal basis vectors in row vectors (10 times magnified)
    rec_plat = np.linalg.inv(plat).T * 10
    symbols = [
        "H",
    ] * len(grgp.qpoints)
    cell = PhonopyAtoms(cell=rec_plat, scaled_positions=grgp.qpoints, symbols=symbols)
    yaml_filename = os.path.join(current_dir, "tio2_qpoints.yaml")
    cell_ref = read_cell_yaml(yaml_filename)
    helper_methods.compare_cells(cell, cell_ref)


def test_length2mesh(ph_nacl: Phonopy):
    """Test of length2mesh."""
    length = 50.0
    mesh_numbers = length2mesh(length, ph_nacl.primitive.cell)
    np.testing.assert_array_equal(mesh_numbers, [15, 15, 15])

    mesh_numbers = length2mesh(
        length, ph_nacl.primitive.cell, ph_nacl.primitive_symmetry.pointgroup_operations
    )
    np.testing.assert_array_equal(mesh_numbers, [15, 15, 15])



import phonopy
from phonopy.structure.cells import isclose

ph_iter = phonopy.load("phonopy_sscha_fc_JPCM2022.yaml.xz", log_level=1)
ph_ha = phonopy.load("phonopy_fc222_JPCM2022.yaml.xz", log_level=1)
ph_ha_444 = phonopy.load("phonopy_fc444_JPCM2022.yaml.xz", log_level=1)

if not isclose(ph_iter.unitcell, ph_ha.unitcell):
    raise RuntimeError("Unitcells of ph_iter and ph_ha are inconsistent.")
if not isclose(ph_iter.unitcell, ph_ha_444.unitcell):
    raise RuntimeError("Unitcells of ph_iter and ph_ha_444 are inconsistent.")

dim = [4, 4, 4]
ipfc_iter_444 = ph_iter.ph2ph(dim).force_constants
ipfc_ha_444 = ph_ha.ph2ph(dim).force_constants
delta_ipfc_444 = ipfc_iter_444 - ipfc_ha_444
ph_ha_444_force_constants = ph_ha_444.force_constants + delta_ipfc_444

ph_ipfc_444 = ph_ha_444.copy()
ph_ipfc_444.force_constants = ph_ha_444_force_constants
ph_ipfc_444.nac_params = ph_iter.nac_params

ph_ipfc_444.save(
    filename="phonopy_ipfc444_JPCM2022.yaml",
    settings={"force_sets": False, "displacements": False, "force_constants": True},
    compression=True,
)



"""Example of band structure and group velocity calculation by NaCl."""

import numpy as np

import phonopy


def _append_band(bands, q_start, q_end):
    band = []
    nq = 51
    for i in range(nq):
        band.append(
            np.array(q_start) + (np.array(q_end) - np.array(q_start)) / (nq - 1) * i
        )
    bands.append(band)


phonon = phonopy.load(
    unitcell_filename="POSCAR-unitcell",
    born_filename="BORN",
    force_sets_filename="FORCE_SETS",
    supercell_matrix=[[2, 0, 0], [0, 2, 0], [0, 0, 2]],
    primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
)

bands = []
_append_band(bands, [0.0, 0.0, 0.0], [0.5, 0.0, 0.0])
_append_band(bands, [0.5, 0.0, 0.0], [0.5, 0.5, 0.0])
_append_band(bands, [0.5, 0.5, 0.0], [0.0, 0.0, 0.0])
_append_band(bands, [0.0, 0.0, 0.0], [0.5, 0.5, 0.5])
phonon.run_band_structure(bands, with_group_velocities=True)
band_dict = phonon.get_band_structure_dict()
frequencies = band_dict["frequencies"]
group_velocities = band_dict["group_velocities"]
print(len(frequencies))
print(frequencies[0].shape)
print(len(group_velocities))
print(group_velocities[0].shape)



"""Group velocity calculation."""

# Copyright (C) 2013 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings
from typing import Optional, Union

import numpy as np

from phonopy.harmonic.derivative_dynmat import DerivativeOfDynamicalMatrix
from phonopy.harmonic.dynamical_matrix import (
    DynamicalMatrix,
    DynamicalMatrixGL,
    DynamicalMatrixNAC,
)
from phonopy.phonon.degeneracy import degenerate_sets
from phonopy.physical_units import get_physical_units
from phonopy.structure.symmetry import Symmetry
from phonopy.utils import similarity_transformation


class GroupVelocity:
    r"""Class to calculate group velocities of phonons.

    d omega   ----
    ------- = \  / omega
    d q        \/q

    Gradient of omega in reciprocal space, which is calculated here by

       1             d D(q)
    ------- <e(q,nu)|------|e(q,nu)>
    2 omega           d q

    Attributes
    ----------
    group_velocity : ndarray
        Group velocities at q-points.
        shape=(q-points, num_band, 3), dtype='double', order='C'
    q_length : float
        Distance in reciprocal space used to calculate finite difference of
        dynamcial matrix.

    """

    Default_q_length = 1e-5

    def __init__(
        self,
        dynamical_matrix: Union[DynamicalMatrix, DynamicalMatrixNAC],
        q_length: Optional[float] = None,
        symmetry: Optional[Symmetry] = None,
        frequency_factor_to_THz: Optional[float] = None,
        cutoff_frequency: float = 1e-4,
    ):
        """Init method.

        dynamical_matrix : DynamicalMatrix or DynamicalMatrixNAC
            Dynamical matrix class instance.
        q_length : float
            This is used such as D(q + q_length) - D(q - q_length) for
            calculating finite difference of dynamical matrix.
            Default is None, which gives 1e-5.
        symmetry : Symmetry
            This is used to symmetrize group velocity at each q-points.
            Default is None, which means no symmetrization.
        frequency_factor_to_THz : float
            Unit conversion factor to convert to THz. Default is VaspToTHz.
        cutoff_frequency : float
            Group velocity is set zero if phonon frequency is below this value.

        """
        self._dynmat = dynamical_matrix
        primitive = dynamical_matrix.primitive
        self._reciprocal_lattice_inv = primitive.cell
        self._reciprocal_lattice = np.linalg.inv(self._reciprocal_lattice_inv)
        self._q_length = q_length
        if isinstance(self._dynmat, DynamicalMatrixGL):
            if self._q_length is None:
                self._q_length = self.Default_q_length

        self._ddm: Optional[DerivativeOfDynamicalMatrix]
        if self._q_length is None:
            self._ddm = DerivativeOfDynamicalMatrix(dynamical_matrix)
        else:
            self._ddm = None

        self._symmetry = symmetry
        if frequency_factor_to_THz is None:
            self._factor = get_physical_units().DefaultToTHz
        else:
            self._factor = frequency_factor_to_THz
        self._cutoff_frequency = cutoff_frequency

        self._directions = np.array(
            [[1, 2, 3], [1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype="double"
        )
        self._directions[0] /= np.linalg.norm(self._directions[0])

        self._q_points = None
        self._group_velocities = None
        self._perturbation = None

    def run(self, q_points, perturbation=None):
        """Group velocities are computed at q-points.

        Calculated group velocities are stored in self._group_velocities.

        Parameters
        ----------
        q_points : array-like
            List of q-points such as [[0, 0, 0], [0.1, 0.2, 0.3], ...].
        perturbation : array-like
            Direction in fractional coordinates of reciprocal space.

        """
        self._q_points = q_points
        self._perturbation = perturbation
        if perturbation is None:
            # Give an random direction to break symmetry
            self._directions[0] = np.array([1, 2, 3])
        else:
            self._directions[0] = np.dot(self._reciprocal_lattice, perturbation)
        self._directions[0] /= np.linalg.norm(self._directions[0])

        gv = [self._calculate_group_velocity_at_q(q) for q in self._q_points]
        self._group_velocities = np.array(gv, dtype="double", order="C")

    @property
    def q_length(self):
        """Setter an getter of q_length."""
        return self._q_length

    @q_length.setter
    def q_length(self, q_length):
        self._q_length = q_length

    def get_q_length(self):
        """Return q_length."""
        warnings.warn(
            "GroupVelocity.get_q_length() is deprecated. "
            "Use q_length attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.q_length

    def set_q_length(self, q_length):
        """Set q_length."""
        warnings.warn(
            "GroupVelocity.set_q_length() is deprecated. "
            "Use q_length attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.q_length = q_length

    @property
    def group_velocities(self):
        """Return group velocities."""
        return self._group_velocities

    def get_group_velocity(self):
        """Return group velocities."""
        warnings.warn(
            "GroupVelocity.get_group_velocity() is deprecated. "
            "Use group_velocities attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.group_velocities

    def _calculate_group_velocity_at_q(self, q):
        self._dynmat.run(q)
        dm = self._dynmat.dynamical_matrix
        eigvals, eigvecs = np.linalg.eigh(dm)
        eigvals = eigvals.real
        freqs = np.sqrt(abs(eigvals)) * np.sign(eigvals) * self._factor
        gv = np.zeros((len(freqs), 3), dtype="double", order="C")
        deg_sets = degenerate_sets(freqs)

        ddms = self._get_dD(np.array(q))
        pos = 0
        for deg in deg_sets:
            gv[pos : pos + len(deg)] = self._perturb_D(ddms, eigvecs[:, deg])
            pos += len(deg)

        for i, f in enumerate(freqs):
            if f > self._cutoff_frequency:
                gv[i, :] *= self._factor**2 / f / 2
            else:
                gv[i, :] = 0

        if self._perturbation is None:
            if self._symmetry is None:
                return gv
            else:
                return self._symmetrize_group_velocity(gv, q)
        else:
            return gv

    def _symmetrize_group_velocity(self, gv, q):
        """Symmetrize obtained group velocities using site symmetries."""
        rotations = []
        for r in self._symmetry.reciprocal_operations:
            q_in_BZ = q - np.rint(q)
            diff = q_in_BZ - np.dot(r, q_in_BZ)
            if (np.abs(diff) < self._symmetry.tolerance).all():
                rotations.append(r)

        gv_sym = np.zeros_like(gv)
        for r in rotations:
            r_cart = similarity_transformation(self._reciprocal_lattice, r)
            gv_sym += np.dot(r_cart, gv.T).T

        return gv_sym / len(rotations)

    def _get_dD(self, q):
        """Compute derivative or finite difference of dynamcial matrices."""
        if self._q_length is None:
            return self._get_dD_analytical(q)
        else:
            return self._get_dD_FD(q)

    def _get_dD_FD(self, q):
        """Compute finite difference of dynamcial matrices."""
        ddm = []
        for dqc in self._directions * self._q_length:
            dq = np.dot(self._reciprocal_lattice_inv, dqc)
            ddm.append(
                _delta_dynamical_matrix(q, dq, self._dynmat) / self._q_length / 2
            )
        return np.array(ddm)

    def _get_dD_analytical(self, q):
        """Compute derivative of dynamcial matrices."""
        self._ddm.run(q)
        ddm = self._ddm.d_dynamical_matrix
        dtype = "c%d" % (np.dtype("double").itemsize * 2)
        ddm_dirs = np.zeros((len(self._directions),) + ddm.shape[1:], dtype=dtype)
        for i, dq in enumerate(self._directions):
            for j in range(3):
                ddm_dirs[i] += dq[j] * ddm[j]
        return ddm_dirs

    def _perturb_D(self, ddms, eigsets):
        """Treat degeneracy.

        Group velocities are calculated using analytical continuation using
        specified directions (self._directions) in reciprocal space.

        ddms : array-like
            List of delta (derivative or finite difference) of dynamical
            matrices along several q-directions for perturbation.
        eigsets : array-like
            List of phonon eigenvectors of degenerate bands.

        """
        _, eigvecs = np.linalg.eigh(np.dot(eigsets.T.conj(), np.dot(ddms[0], eigsets)))

        gv = []
        rot_eigsets = np.dot(eigsets, eigvecs)
        for ddm in ddms[1:]:
            gv.append(
                np.diag(np.dot(rot_eigsets.T.conj(), np.dot(ddm, rot_eigsets))).real
            )

        return np.transpose(gv)


def get_group_velocity(
    q,  # q-point
    dynamical_matrix,
    q_length=None,  # finite distance in q
    symmetry=None,
    frequency_factor_to_THz=None,
):
    """Return group velocity at a q-point."""
    if frequency_factor_to_THz is None:
        _factor = get_physical_units().DefaultToTHz
    else:
        _factor = frequency_factor_to_THz
    gv = GroupVelocity(
        dynamical_matrix,
        q_length=q_length,
        symmetry=symmetry,
        frequency_factor_to_THz=_factor,
    )
    gv.run([q])
    return gv.group_velocity[0]


def _delta_dynamical_matrix(q, delta_q, dynmat):
    dynmat.run(q - delta_q)
    dm1 = dynmat.dynamical_matrix
    dynmat.run(q + delta_q)
    dm2 = dynmat.dynamical_matrix
    return dm2 - dm1



"""Tests for PWmat calculater interface."""

import os

import numpy as np

from phonopy.interface.phonopy_yaml import read_cell_yaml
from phonopy.interface.pwmat import read_atom_config
from phonopy.structure import cells

data_dir = os.path.dirname(os.path.abspath(__file__))


def test_read_pwmat():
    """Test of read_PWmat."""
    cell = read_atom_config(os.path.join(data_dir, "Si-pwmat.config"))
    filename = os.path.join(data_dir, "Si-pwmat.yaml")
    cell_ref = read_cell_yaml(filename)
    assert cells.isclose(cell, cell_ref)
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r


def test_magmom():
    """Test of read_PWmat_magmom."""
    cell = read_atom_config(os.path.join(data_dir, "Si-pwmat.config"))
    filename = os.path.join(data_dir, "Si-pwmat.yaml")
    cell_ref = read_cell_yaml(filename)
    assert cells.isclose(cell, cell_ref)
    diff_mag = cell_ref.magnetic_moments - np.array([1] * 8)
    assert (np.abs(diff_mag) < 1e-5).all()



"""Symfc force constants calculator interface."""

# Copyright (C) 2024 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import math
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np

from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive
from phonopy.structure.symmetry import Symmetry


def run_symfc(
    supercell: PhonopyAtoms,
    primitive: Primitive,
    displacements: np.ndarray,
    forces: np.ndarray,
    orders: Optional[Sequence[int]] = None,
    is_compact_fc: bool = False,
    symmetry: Optional[Symmetry] = None,
    options: Optional[Union[str, dict]] = None,
    log_level: int = 0,
) -> dict[int, np.ndarray]:
    """Calculate force constants using symfc.

    The details of the parameters are found in the SymfcFCSolver class.

    """
    if orders is None:
        _orders = [2]
    else:
        _orders = orders

    options_dict = parse_symfc_options(options)

    symfc_calculator = SymfcFCSolver(
        supercell,
        displacements=displacements,
        forces=forces,
        symmetry=symmetry,
        orders=_orders,
        options=options_dict,
        is_compact_fc=is_compact_fc,
        log_level=log_level,
    )
    assert np.array_equal(symfc_calculator.p2s_map, primitive.p2s_map)

    return symfc_calculator.force_constants


class SymfcFCSolver:
    """Interface to symfc."""

    def __init__(
        self,
        supercell: PhonopyAtoms,
        displacements: Optional[np.ndarray] = None,
        forces: Optional[np.ndarray] = None,
        symmetry: Optional[Symmetry] = None,
        orders: Optional[Sequence[int]] = None,
        options: Optional[dict] = None,
        is_compact_fc: bool = False,
        log_level: int = 0,
    ):
        """Init method.

        Parameters
        ----------
        supercell : PhonopyAtoms
            Supercell.
        displacements : np.ndarray, optional
            Displacements. Default is None.
        forces : np.ndarray, optional.
            Forces. Default is None.
        symmetry : Symmetry, optional
            Symmetry of supercell. Default is None.
        orders: Sequence[int], optional
            Orders of force constants. Default is None.
        options : dict, optional
            Options for symfc. Default is None, which gives {}.
        is_compact_fc : bool, optional
            Whether force constants are compact or full. When True, check if
            SymfcFCSolver.p2s_map is equal to primitive.p2s_map. Default is
            False.
        log_level : int, optional
            Log level. Default is 0.

        """
        if options is None:
            self._options = {}
        else:
            self._options = options
        self._supercell = supercell
        self._log_level = log_level
        self._orders = orders
        self._is_compact_fc = is_compact_fc
        self._symfc = None

        self._initialize(
            supercell,
            symmetry=symmetry,
            displacements=displacements,
            forces=forces,
        )
        if self._orders is not None:
            self.run(self._orders)

    @property
    def force_constants(self) -> dict[int, np.ndarray]:
        """Return force constants.

        Returns
        -------
        dict[int, np.ndarray]
            Force constants with order as key.

        """
        if self._orders is None:
            raise RuntimeError("Run SymfcCalculator.run() first.")
        return self._symfc.force_constants

    def run(self, orders: Sequence[int]):
        """Run symfc."""
        if self._log_level:
            print(
                "--------------------------------"
                " Symfc start "
                "-------------------------------"
            )
            print("Symfc is a force constants calculator. See the following paper:")
            print("A. Seko and A. Togo, Phys. Rev. B, 110, 214302 (2024).")
            print("Symfc is developed at https://github.com/symfc/symfc.")
            print(f"Computing {orders} order force constants.", flush=True)
            if self._options:
                print("Parameters:")
                for key, val in self._options.items():
                    print(f"  {key}: {val}", flush=True)

        self._orders = orders
        self._symfc.run(orders=orders, is_compact_fc=self._is_compact_fc)

        if self._log_level == 1:
            print("Increase log-level to watch detailed symfc log.")

        if self._log_level:
            print(
                "---------------------------------"
                " Symfc end "
                "--------------------------------"
            )

    @property
    def p2s_map(self) -> np.ndarray:
        """Return indices of translationally independent atoms."""
        return self._symfc.p2s_map

    @property
    def basis_set(self) -> dict:
        """Return basis set."""
        return self._symfc.basis_set

    def estimate_basis_size(
        self,
        max_order: Optional[int] = None,
        orders: Optional[list] = None,
    ) -> dict:
        """Estimate basis size."""
        basis_sizes = self._symfc.estimate_basis_size(
            orders=orders, max_order=max_order
        )
        return basis_sizes

    def estimate_numbers_of_supercells(
        self,
        max_order: Optional[int] = None,
        orders: Optional[list] = None,
    ) -> dict:
        """Estimate numbers of supercells."""
        basis_sizes = self.estimate_basis_size(max_order=max_order, orders=orders)
        n_scells = {}
        for order, basis_size in basis_sizes.items():
            n_scells[order] = math.ceil(basis_size / len(self._symfc.supercell) / 3)
        return n_scells

    def _initialize(
        self,
        supercell: PhonopyAtoms,
        symmetry: Optional[Symmetry] = None,
        displacements: Optional[np.ndarray] = None,
        forces: Optional[np.ndarray] = None,
    ):
        """Calculate force constants."""
        try:
            from symfc import Symfc
            from symfc.utils.utils import SymfcAtoms
        except ImportError as exc:
            raise ModuleNotFoundError("Symfc python module was not found.") from exc

        symfc_supercell = SymfcAtoms(
            cell=supercell.cell,
            scaled_positions=supercell.scaled_positions,
            numbers=supercell.numbers,
        )
        spacegroup_operations = symmetry.symmetry_operations if symmetry else None
        self._symfc = Symfc(
            symfc_supercell,
            spacegroup_operations=spacegroup_operations,
            cutoff=self._options.get("cutoff"),
            use_mkl=self._options.get("use_mkl", False),
            log_level=self._log_level - 1 if self._log_level else 0,
        )
        if displacements is not None and forces is not None:
            self._symfc.displacements = displacements
            self._symfc.forces = forces

    def compute_basis_set(
        self,
        max_order: Optional[int] = None,
        orders: Optional[list] = None,
    ) -> dict:
        """Run basis set calculations and return basis sets.

        Parameters
        ----------
        max_order : int
            Maximum fc order.
        orders: list
            Orders of force constants.

        Returns
        -------
        dict[FCBasisSetBase]
            Basis sets. Keys are orders.

        """
        self._symfc.compute_basis_set(max_order=max_order, orders=orders)
        return self._symfc.basis_set


def parse_symfc_options(options: Optional[Union[str, dict]]) -> dict:
    """Parse symfc options.

    Parameters
    ----------
    options : Union[str, dict]
        Options for symfc.

    Returns
    -------
    dict

    Note
    ----
    When str, it should be written as follows:

        "cutoff = 10.0"

    """
    if options is None:
        return {}
    if isinstance(options, dict):
        return options
    elif isinstance(options, str):
        options_dict = {}
        for option in options.split(","):
            key_val = [v.strip().lower() for v in option.split("=")]
            if len(key_val) != 2:
                break
            key, val = key_val
            if key == "cutoff":
                try:
                    options_dict[key] = {3: float(val)}
                except ValueError:
                    print("Warning: Cutoff value must be float.")
            if key == "use_mkl":
                if val.strip().lower() == "true":
                    options_dict[key] = True
        return options_dict
    else:
        raise TypeError(f"options must be str or dict, not {type(options)}.")



"""Crystal symmetry routines."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import warnings
from collections.abc import Sequence
from typing import Optional

try:
    from spglib import SpglibDataset
except ImportError:
    from types import SimpleNamespace as SpglibDataset

import numpy as np
import spglib

from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import (
    compute_all_sg_permutations,
    get_primitive,
    get_supercell,
)
from phonopy.utils import get_dot_access_dataset, similarity_transformation


class Symmetry:
    """Class to find and store crystal symmetry information."""

    def __init__(
        self,
        cell: PhonopyAtoms,
        symprec: float = 1e-5,
        is_symmetry: bool = True,
        s2p_map: Optional[np.ndarray] = None,
    ):
        """Init method.

        cell : PhonopyAtoms
            Crystal structure whose symmetry is analyzed.
        symprec : float, optional
            Tolerance used to find crystal symmetry. Default is 1e-5.
        is_symmetry : bool, optional
            With or without symmetry analysis. When `is_symmetry=False` and
            `p2s_map` is given, pure translations inside `cell` is registered
            as symmetry operations. Default is True.
        s2p_map : ndarray, optional
            This is equivalent to `Primitive.s2p_map`.

        """
        self._cell = cell
        self._symprec = symprec

        self._symmetry_operations: dict
        self._international_table = None
        self._dataset = None
        self._wyckoff_letters = None
        self._map_atoms = None
        self._atomic_permutations = None
        self._pointgroup_operations = None
        self._pointgroup = None
        self._independent_atoms = None
        self._map_operations = None

        magmom = cell.magnetic_moments

        if not is_symmetry:
            self._set_nosym(s2p_map)
        elif magmom is None:
            self._set_symmetry_dataset()
        else:
            self._set_symmetry_operations_with_magmoms()
        (
            self._pointgroup_operations,
            self._reciprocal_operations,
        ) = get_pointgroup_operations(self._symmetry_operations["rotations"])
        ptg_symbol = spglib.get_pointgroup(self._pointgroup_operations)[0]
        self._pointgroup = ptg_symbol.strip()
        self._set_atomic_permutations()
        self._set_independent_atoms()
        self._map_operations = self._get_map_operations_from_permutations()

    @property
    def symmetry_operations(self):
        """Return symmetry operations.

        Returns
        -------
        dict
            'rotations': ndarray
                Matrix parts.
                shape(num_operations, 3, 3), dtype='intc',
             'translations': ndarray
                Vector parts.
                shape(num_operations, 3), dtype='double',

        """
        return self._symmetry_operations

    def get_symmetry_operations(self):
        """Return symmetry operations."""
        warnings.warn(
            "Symmetry.get_symmetry_operations() is deprecated."
            "Use symmetry_operations attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.symmetry_operations

    def get_symmetry_operation(self, operation_number):
        """Return one symmetry operation."""
        operation = self._symmetry_operations
        return {
            "rotations": operation["rotations"][operation_number],
            "translations": operation["translations"][operation_number],
        }

    @property
    def pointgroup_operations(self):
        """Return crystallographic point group operations."""
        return self._pointgroup_operations

    def get_pointgroup_operations(self):
        """Return crystallographic point group operations."""
        warnings.warn(
            "Symmetry.get_pointgroup_operations() is deprecated."
            "Use pointgroup_operations attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.pointgroup_operations

    @property
    def pointgroup_symbol(self):
        """Return symbol of crystallographic point group."""
        return self._pointgroup

    def get_pointgroup(self):
        """Return symbol of crystallographic point group."""
        warnings.warn(
            "Symmetry.get_pointgroup() is deprecated.Use pointgroup_symbol attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._pointgroup

    def get_international_table(self):
        """Return international symbol of space group."""
        return self._international_table

    def get_Wyckoff_letters(self):
        """Return Wycloff letters."""
        return self._wyckoff_letters

    @property
    def dataset(self) -> SpglibDataset:
        """Return spglib dataset.

        This is raw data of symmetry.

        """
        return self._dataset

    def get_dataset(self):
        """Return spglib dataset."""
        warnings.warn(
            "Symmetry.get_dataset() is deprecated.Use dataset attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.dataset

    def get_independent_atoms(self):
        """Return symmetrically unique atoms."""
        return self._independent_atoms

    def get_map_atoms(self):
        """Return equivalent_atoms of spglib dataset.

        Returns
        -------
        spglib_dataset.equivalent_atoms
        e.g.,
            [0, 0, 0, 0, 4, 4, 4, 4]

        """
        return self._map_atoms

    def get_map_operations(self):
        """Return symmetry operation indices that map to respective equivalent atoms.

        Returns
        -------
        operations : ndarray
            Indices of symmetry operations that sent atoms to respective
            equivalent atoms. For each atom, only one of those symmetry
            operations is stored. When all symmetry mapping information
            is needed, ``atomic_permulations`` is used.
            shape=(atoms,), dtype='intc'

        """
        return self._map_operations

    def get_site_symmetry(self, atom_number):
        """Return matrix parts of site symmetry operations."""
        positions = self._cell.scaled_positions
        lattice = self._cell.cell
        rotations = self._symmetry_operations["rotations"]
        translations = self._symmetry_operations["translations"]

        return self._get_site_symmetry(
            atom_number, lattice, positions, rotations, translations, self._symprec
        )

    @property
    def tolerance(self):
        """Return symmetry tolerance."""
        return self._symprec

    def get_symmetry_tolerance(self):
        """Return symmetry tolerance."""
        warnings.warn(
            "Symmetry.get_symmetry_tolerance() is deprecated."
            "Use tolerance attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.tolerance

    @property
    def reciprocal_operations(self):
        """Return reciprocal space point group operations.

        Definition of operation:
        q' = Rq

        This is transpose of that shown in ITA (q' = qR).

        """
        return self._reciprocal_operations

    def get_reciprocal_operations(self):
        """Return reciprocal space point group operations."""
        warnings.warn(
            "Symmetry.get_reciprocal_operations() is deprecated."
            "Use reciprocal_operations attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.reciprocal_operations

    @property
    def atomic_permutations(self):
        """Return atomic index permutations by space group operations.

        shape=(operations, positions)

        See compute_all_sg_permutations.

        """
        return self._atomic_permutations

    def get_atomic_permutations(self):
        """Return atomic index permutations by space group operations."""
        warnings.warn(
            "Symmetry.get_atomic_permutations() is deprecated."
            "Use atomic_permutations attribute instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.atomic_permutations

    def _set_atomic_permutations(self):
        positions = self._cell.scaled_positions
        lattice = np.array(self._cell.cell.T, dtype="double", order="C")
        rotations = self._symmetry_operations["rotations"]
        translations = self._symmetry_operations["translations"]
        self._atomic_permutations = compute_all_sg_permutations(
            positions,  # scaled positions
            rotations,  # scaled
            translations,  # scaled
            lattice,  # column vectors
            self._symprec,
        )

    def _get_site_symmetry(
        self, atom_number, lattice, positions, rotations, translations, symprec
    ):
        pos = positions[atom_number]
        site_symmetries = []

        for r, t in zip(rotations, translations):
            rot_pos = np.dot(pos, r.T) + t
            diff = pos - rot_pos
            diff -= np.rint(diff)
            diff = np.dot(diff, lattice)
            if np.linalg.norm(diff) < symprec:
                site_symmetries.append(r)

        return np.array(site_symmetries, dtype="intc")

    def _set_symmetry_dataset(self):
        self._dataset = get_dot_access_dataset(
            spglib.get_symmetry_dataset(self._cell.totuple(), self._symprec)
        )

        self._symmetry_operations = {
            "rotations": self._dataset.rotations,
            "translations": self._dataset.translations,
        }
        self._international_table = "%s (%d)" % (
            self._dataset.international,
            self._dataset.number,
        )
        self._wyckoff_letters = self._dataset.wyckoffs

        self._map_atoms = self._dataset.equivalent_atoms

    def _set_symmetry_operations_with_magmoms(self):
        self._dataset = get_dot_access_dataset(
            spglib.get_magnetic_symmetry_dataset(
                self._cell.totuple(), symprec=self._symprec
            )
        )
        self._symmetry_operations = {
            "rotations": self._dataset.rotations,
            "translations": self._dataset.translations,
        }
        self._map_atoms = self._dataset.equivalent_atoms

    def _set_independent_atoms(self):
        indep_atoms = []
        for i, atom_map in enumerate(self._map_atoms):
            if i == atom_map:
                indep_atoms.append(i)
        self._independent_atoms = np.array(indep_atoms, dtype="intc")

    def _set_map_operations(self):
        warnings.warn(
            "Symmetry._set_map_operations is deprecated."
            "This was replaced by "
            "_get_map_operations_from_permutations.",
            DeprecationWarning,
            stacklevel=2,
        )

        ops = self._symmetry_operations
        pos = self._cell.scaled_positions
        lattice = self._cell.cell
        map_operations = np.zeros(len(pos), dtype="intc")

        for i, eq_atom in enumerate(self._map_atoms):
            for j, (r, t) in enumerate(zip(ops["rotations"], ops["translations"])):
                diff = np.dot(pos[i], r.T) + t - pos[eq_atom]
                diff -= np.rint(diff)
                dist = np.linalg.norm(np.dot(diff, lattice))
                if dist < self._symprec:
                    map_operations[i] = j
                    break
        self._map_operations = map_operations

    def _get_map_operations_from_permutations(self):
        perm = self._atomic_permutations
        map_operations = np.zeros(perm.shape[1], dtype="intc")
        for i, eq_atom in enumerate(self._map_atoms):
            match = np.where(perm[:, i] == eq_atom)[0]
            assert len(match) != 0
            map_operations[i] = match[0]
        return map_operations

    def _set_nosym(self, s2p_map):
        translations = []
        rotations = []

        if s2p_map is None:
            rotations.append(np.eye(3, dtype="intc"))
            translations.append(np.zeros(3, dtype="double"))
            self._map_atoms = range(len(self._cell))
        else:
            positions = self._cell.scaled_positions
            for i, j in enumerate(s2p_map):
                if j == 0:
                    ipos0 = i
                    break
            for i, p in zip(s2p_map, positions):
                if i == 0:
                    trans = p - positions[ipos0]
                    trans -= np.floor(trans)
                    translations.append(trans)
                    rotations.append(np.eye(3, dtype="intc"))
            self._map_atoms = s2p_map
        self._symmetry_operations = {
            "rotations": np.array(rotations, dtype="intc"),
            "translations": np.array(translations, dtype="double"),
        }
        self._international_table = "P1 (1)"
        self._wyckoff_letters = ["a"] * len(self._cell)


def get_pointgroup_operations(rotations, is_time_reversal=True):
    """Return direct and reciprocal point group operations."""
    ptg_ops = collect_unique_rotations(rotations)
    reciprocal_rotations = [rot.T for rot in ptg_ops]

    if is_time_reversal:
        exist_r_inv = False
        for rot in ptg_ops:
            if (rot == -np.eye(3, dtype="intc")).all():
                exist_r_inv = True
                break
        if not exist_r_inv:
            reciprocal_rotations += [-rot.T for rot in ptg_ops]

    return (
        np.array(ptg_ops, dtype="intc"),
        np.array(reciprocal_rotations, dtype="intc"),
    )


def collect_unique_rotations(rotations):
    """Collect unique rotations in input rotation matrices."""
    ptg_ops = []
    for rot in rotations:
        is_same = False
        for tmp_rot in ptg_ops:
            if (tmp_rot == rot).all():
                is_same = True
                break
        if not is_same:
            ptg_ops.append(rot)

    return ptg_ops


def get_lattice_vector_equivalence(point_symmetry):
    """Return equivalences of basis vector length pairs, (b==c, c==a, a==b).

    Change of basis is defined by

    (a', b', c') = (a, b, c) R.

    Then, check rotation matrices if
        a -> b, b -> c, c -> a, a -> -b, b -> -c, and c -> -a,
    can happen.

    Parameters
    ----------
    point_symmetry : array_like
        Rotation matrices.
        shape=(n_rot, 3, 3), dtype=int.

    """
    # primitive_vectors: column vectors

    equivalence = [False, False, False]
    for r in point_symmetry:
        if (np.abs(r[:, 0]) == [0, 1, 0]).all():
            equivalence[2] = True
        if (np.abs(r[:, 0]) == [0, 0, 1]).all():
            equivalence[1] = True
        if (np.abs(r[:, 1]) == [1, 0, 0]).all():
            equivalence[2] = True
        if (np.abs(r[:, 1]) == [0, 0, 1]).all():
            equivalence[0] = True
        if (np.abs(r[:, 2]) == [1, 0, 0]).all():
            equivalence[1] = True
        if (np.abs(r[:, 2]) == [0, 1, 0]).all():
            equivalence[0] = True

    return equivalence


def elaborate_borns_and_epsilon(
    ucell,
    borns,
    epsilon,
    primitive_matrix=None,
    supercell_matrix=None,
    is_symmetry=True,
    symmetrize_tensors=False,
    symprec=1e-5,
):
    """Symmetrize Born effective charges and dielectric constants.

    Born effective charges of symmetrically independent atoms
    for primitive cell are extracted.


    Parameters
    ----------
    ucell : PhonopyAtoms
        Unit cell structure
    borns : array_like
        Born effective charges of ucell
    epsilon : array_like
        Dielectric constant tensor

    Returns
    -------
    ndarray
        Born effective charges of symmetrically independent atoms in primitive cell
        Dielectric constant
    ndarray
        Atomic index mapping table from supercell to primitive cell
        of independent atoms

    Raises
    ------
    AssertionError
        Inconsistency of number of atoms or Born effective charges.

    Warning
    -------
    Broken symmetry of Born effective charges

    """
    assert len(borns) == len(ucell), "num_atom %d != len(borns) %d" % (
        len(ucell),
        len(borns),
    )

    if symmetrize_tensors:
        borns_, epsilon_ = symmetrize_borns_and_epsilon(
            borns, epsilon, ucell, symprec=symprec, is_symmetry=is_symmetry
        )
    else:
        borns_ = borns
        epsilon_ = epsilon

    indeps_in_supercell, indeps_in_unitcell = _extract_independent_atoms(
        ucell,
        primitive_matrix=primitive_matrix,
        supercell_matrix=supercell_matrix,
        is_symmetry=is_symmetry,
        symprec=symprec,
    )

    return borns_[indeps_in_unitcell].copy(), epsilon_, indeps_in_supercell


def symmetrize_borns_and_epsilon(
    borns: Sequence,
    epsilon: Sequence,
    ucell: PhonopyAtoms,
    primitive_matrix: Optional[Sequence] = None,
    primitive: PhonopyAtoms = None,
    supercell_matrix: Optional[Sequence] = None,
    symprec: float = 1e-5,
    is_symmetry: bool = True,
):
    """Symmetrize Born effective charges and dielectric tensor.

    Parameters
    ----------
    borns: array_like
        Born effective charges.
        shape=(unitcell_atoms, 3, 3)
        dtype='double'
    epsilon: array_like
        Dielectric constant
        shape=(3, 3)
        dtype='double'
    ucell: PhonopyAtoms
        Unit cell
    primitive_matrix: array_like, optional
        Primitive matrix. This is used to select Born effective charges in
        primitive cell. If None (default), Born effective charges in unit cell
        are returned.
        shape=(3, 3)
        dtype='double'
    primitive : PhonopyAtoms
        This is an alternative of giving primitive_matrix (Mp). Mp is given as
            Mp = (a_u, b_u, c_u)^-1 * (a_p, b_p, c_p).
        In addition, the order of atoms is alined to those of atoms in this
        primitive cell for Born effective charges. No rigid rotation of
        crystal structure is assumed.
    supercell_matrix: array_like, optional
        Supercell matrix. This is used to select Born effective charges in
        **primitive cell**. Supercell matrix is needed because primitive
        cell is created first creating supercell from unit cell, then
        the primitive cell is created from the supercell. If None (defautl),
        1x1x1 supercell is created.
        shape=(3, 3)
        dtype='int'
    symprec: float, optional
        Symmetry tolerance. Default is 1e-5
    is_symmetry: bool, optinal
        By setting False, symmetrization can be switched off. Default is True.

    """
    lattice = ucell.cell
    u_sym = Symmetry(ucell, is_symmetry=is_symmetry, symprec=symprec)
    rotations = u_sym.symmetry_operations["rotations"]
    translations = u_sym.symmetry_operations["translations"]
    ptg_ops = u_sym.pointgroup_operations
    epsilon_ = _symmetrize_2nd_rank_tensor(epsilon, ptg_ops, lattice)
    borns_ = _take_average_of_borns(borns, rotations, translations, ucell, symprec)

    if (abs(borns - borns_) > 0.1).any():
        lines = [
            "Symmetry of Born effective charge is largely broken. The difference is:",
            "%s" % (borns - borns_),
        ]
        import warnings

        warnings.warn("\n".join(lines), stacklevel=2)

    if primitive_matrix is None and primitive is None:
        return borns_, epsilon_
    else:
        if primitive is not None:
            pmat = np.dot(np.linalg.inv(ucell.cell.T), primitive.cell.T)
        else:
            pmat = primitive_matrix

        scell, pcell = _get_supercell_and_primitive(
            ucell,
            primitive_matrix=pmat,
            supercell_matrix=supercell_matrix,
            symprec=symprec,
        )

        idx = [scell.u2u_map[i] for i in scell.s2u_map[pcell.p2s_map]]
        borns_in_prim = borns_[idx].copy()

        if primitive is None:
            return borns_in_prim, epsilon_
        else:
            idx2 = _get_mapping_between_cells(pcell, primitive)
            return borns_in_prim[idx2].copy(), epsilon_


def _take_average_of_borns(
    borns: np.ndarray,
    rotations: np.ndarray,
    translations: np.ndarray,
    cell: PhonopyAtoms,
    symprec: float,
) -> np.ndarray:
    lattice = cell.cell
    positions = cell.scaled_positions
    borns_ = np.zeros_like(borns)
    for i in range(len(borns)):
        for r, t in zip(rotations, translations):
            diff = np.dot(positions, r.T) + t - positions[i]
            diff -= np.rint(diff)
            dist = np.sqrt(np.sum(np.dot(diff, lattice) ** 2, axis=1))
            j = np.nonzero(dist < symprec)[0][0]
            r_cart = similarity_transformation(lattice.T, r)
            borns_[i] += similarity_transformation(r_cart, borns[j])
        borns_[i] /= len(rotations)

    sum_born = borns_.sum(axis=0) / len(borns_)
    borns_ -= sum_born

    return borns_


def _get_mapping_between_cells(
    cell_from: PhonopyAtoms, cell_to: PhonopyAtoms, symprec: float = 1e-5
):
    indices = []
    lattice = cell_from.cell
    pos_from = cell_from.scaled_positions
    for p_to in cell_to.scaled_positions:
        diff = pos_from - p_to
        diff -= np.rint(diff)
        dist = np.sqrt(np.sum(np.dot(diff, lattice) ** 2, axis=1))
        ids = np.nonzero(dist < symprec)[0]
        if len(ids) == 1:
            indices.append(ids[0])
        else:
            msg = "Index matching didn't go well."
            raise RuntimeError(msg)
    return indices


def _symmetrize_2nd_rank_tensor(tensor, symmetry_operations, lattice):
    sym_cart = [similarity_transformation(lattice.T, r) for r in symmetry_operations]
    sum_tensor = np.zeros_like(tensor)
    for sym in sym_cart:
        sum_tensor += similarity_transformation(sym, tensor)
    return sum_tensor / len(symmetry_operations)


def _extract_independent_atoms(
    ucell, primitive_matrix=None, supercell_matrix=None, is_symmetry=True, symprec=1e-5
):
    scell, pcell = _get_supercell_and_primitive(
        ucell,
        primitive_matrix=primitive_matrix,
        supercell_matrix=supercell_matrix,
        symprec=symprec,
    )
    p_sym = Symmetry(pcell, is_symmetry=is_symmetry, symprec=symprec)
    s_indep_atoms = pcell.p2s_map[p_sym.get_independent_atoms()]
    u_indep_atoms = [scell.u2u_map[x] for x in s_indep_atoms]

    return s_indep_atoms, u_indep_atoms


def _get_supercell_and_primitive(
    ucell, primitive_matrix=None, supercell_matrix=None, symprec=1e-5
):
    if primitive_matrix is None:
        pmat = np.eye(3)
    else:
        pmat = primitive_matrix
    if supercell_matrix is None:
        smat = np.eye(3, dtype="intc")
    else:
        smat = supercell_matrix

    inv_smat = np.linalg.inv(smat)
    scell = get_supercell(ucell, smat, symprec=symprec)
    pcell = get_primitive(scell, np.dot(inv_smat, pmat), symprec=symprec)

    return scell, pcell



"""Tests of VASP calculator interface."""

import tarfile
import tempfile
from io import StringIO
from pathlib import Path

import numpy as np
import pytest

from phonopy.file_IO import parse_FORCE_SETS
from phonopy.interface.phonopy_yaml import read_cell_yaml
from phonopy.interface.vasp import (
    Vasprun,
    VasprunxmlExpat,
    get_vasp_structure_lines,
    parse_set_of_forces,
    read_vasp,
    read_vasp_from_strings,
    read_XDATCAR,
    write_XDATCAR,
)

cwd = Path(__file__).parent


def test_read_vasp():
    """Test read_vasp."""
    cell = read_vasp(cwd / ".." / "POSCAR_NaCl")
    filename = cwd / "NaCl-vasp.yaml"
    cell_ref = read_cell_yaml(filename)
    assert (np.abs(cell.cell - cell_ref.cell) < 1e-5).all()
    diff_pos = cell.scaled_positions - cell_ref.scaled_positions
    diff_pos -= np.rint(diff_pos)
    assert (np.abs(diff_pos) < 1e-5).all()
    for s, s_r in zip(cell.symbols, cell_ref.symbols):
        assert s == s_r


@pytest.mark.parametrize(
    "first_line_str, is_vasp4",
    [(None, True), (None, False), ("my_comment", True), ("my_comment", False)],
)
def test_get_vasp_structure_lines(helper_methods, is_vasp4, first_line_str):
    """Test get_vasp_structure_lines (almost write_vasp)."""
    filename = cwd / "NaCl-vasp.yaml"
    cell_ref = read_cell_yaml(filename)
    lines = get_vasp_structure_lines(
        cell_ref, direct=True, is_vasp4=is_vasp4, first_line_str=first_line_str
    )
    cell = read_vasp_from_strings("\n".join(lines))
    helper_methods.compare_cells_with_order(cell, cell_ref)
    if is_vasp4:
        # is_vasp4 is True, first_line_str is ignored.
        assert lines[0] == "Na Cl"
    elif first_line_str:
        assert lines[0] == first_line_str


def test_get_vasp_structure_lines_shuffled_positions(helper_methods):
    """Test get_vasp_structure_lines with a cell having shuffled positions.

    Order of atoms is sorted by chemical symbols. Therefore,
    helper_methods.compare_cells_with_order fails.

    """
    poscar_yaml = """lattice:
- [     5.690301476175672,     0.000000000000000,     0.000000000000000 ] # a
- [     0.000000000000000,     5.690301476175672,     0.000000000000000 ] # b
- [     0.000000000000000,     0.000000000000000,     5.690301476175672 ] # c
points:
- symbol: Na # 1
  coordinates: [  0.000000000000000,  0.000000000000000,  0.000000000000000 ]
  mass: 22.989769
- symbol: Cl # 2
  coordinates: [  0.500000000000000,  0.000000000000000,  0.000000000000000 ]
  mass: 35.453000
- symbol: Na # 3
  coordinates: [  0.000000000000000,  0.500000000000000,  0.500000000000000 ]
  mass: 22.989769
- symbol: Cl # 4
  coordinates: [  0.500000000000000,  0.500000000000000,  0.500000000000000 ]
  mass: 35.453000
- symbol: Na # 5
  coordinates: [  0.500000000000000,  0.000000000000000,  0.500000000000000 ]
  mass: 22.989769
- symbol: Cl # 6
  coordinates: [  0.000000000000000,  0.000000000000000,  0.500000000000000 ]
  mass: 35.453000
- symbol: Na # 7
  coordinates: [  0.500000000000000,  0.500000000000000,  0.000000000000000 ]
  mass: 22.989769
- symbol: Cl # 8
  coordinates: [  0.000000000000000,  0.500000000000000,  0.000000000000000 ]
  mass: 35.453000
  """
    cell_ref = read_cell_yaml(StringIO(poscar_yaml))
    lines = get_vasp_structure_lines(cell_ref, direct=True)
    cell = read_vasp_from_strings("\n".join(lines))
    with pytest.raises(AssertionError):
        helper_methods.compare_cells_with_order(cell, cell_ref)
    helper_methods.compare_cells(cell, cell_ref)
    perm = []
    for p_i in cell.scaled_positions:
        diff = cell_ref.scaled_positions - p_i
        diff -= np.rint(diff)
        dists = np.linalg.norm(diff @ cell.cell, axis=1)
        perm.append(np.where(dists < 1e-8)[0][0])
    np.testing.assert_array_equal(perm, [0, 2, 4, 6, 1, 3, 5, 7])
    np.testing.assert_array_equal(cell.numbers, [11, 11, 11, 11, 17, 17, 17, 17])


def test_parse_vasprun_xml():
    """Test parsing vasprun.xml with expat."""
    filename_vasprun = cwd / "vasprun.xml.tar.bz2"
    _tar = tarfile.open(filename_vasprun)
    filename = cwd / ".." / "FORCE_SETS_NaCl"
    dataset = parse_FORCE_SETS(filename=filename)
    energy_ref = [-216.82820693, -216.82817843]
    for i, member in enumerate(_tar.getmembers()):
        vr = Vasprun(_tar.extractfile(member), use_expat=True)
        # for force in vr.read_forces():
        #     print("% 15.8f % 15.8f % 15.8f" % tuple(force))
        # print("")
        ref = dataset["first_atoms"][i]["forces"]
        np.testing.assert_allclose(ref, vr.read_forces(), atol=1e-8)
        np.testing.assert_allclose(energy_ref[i], vr.read_energy(), atol=1e-8)


def test_VasprunxmlExpat():
    """Test VasprunxmlExpat."""
    filename_vasprun = cwd / "vasprun.xml.tar.bz2"
    _tar = tarfile.open(filename_vasprun)
    for member in _tar.getmembers():
        vasprun = VasprunxmlExpat(_tar.extractfile(member))
        vasprun.parse()
        np.testing.assert_equal(vasprun.fft_grid, [64, 64, 64])
        np.testing.assert_equal(vasprun.fft_fine_grid, [128, 128, 128])
        assert vasprun.efermi is None
        assert vasprun.symbols == ["Na"] * 32 + ["Cl"] * 32
        np.testing.assert_almost_equal(vasprun.NELECT, 448)
        np.testing.assert_almost_equal(vasprun.volume, 1473.99433936)
        break


def test_parse_set_of_forces():
    """Test parse_set_of_forces."""
    filename_vasprun = cwd / "vasprun.xml.tar.bz2"
    _tar = tarfile.open(filename_vasprun)
    fps = [_tar.extractfile(member) for member in _tar.getmembers()]
    calc_dataset = parse_set_of_forces(64, fps)
    filename = cwd / ".." / "FORCE_SETS_NaCl"
    dataset = parse_FORCE_SETS(filename=filename)
    force_sets = [dataset["first_atoms"][i]["forces"] for i in (0, 1)]
    energy_ref = [-216.82820693, -216.82817843]
    np.testing.assert_allclose(
        calc_dataset["points"][0][0], [0.00087869, 0.0, 0.0], atol=1e-5
    )
    np.testing.assert_allclose(
        calc_dataset["points"][1][32], [0.25087869, 0.25, 0.25], atol=1e-5
    )
    np.testing.assert_allclose(force_sets, calc_dataset["forces"], atol=1e-8)
    np.testing.assert_allclose(
        energy_ref, calc_dataset["supercell_energies"], atol=1e-8
    )


def test_read_XDATCAR():
    """Test read_XDATCAR."""
    filename_xdatcar = cwd / "XDATCAR-NaCl"
    lattice, positions = read_XDATCAR(filename_xdatcar)

    np.testing.assert_allclose(lattice, np.eye(3) * 22.562240, atol=1e-8)
    np.testing.assert_allclose(
        positions[0, 0], [0.99854664, 0.00265936, 0.00701660], atol=1e-8
    )
    np.testing.assert_allclose(
        positions[-1, -1], [0.75034458, 0.74972945, 0.87262656], atol=1e-8
    )


def test_write_XDATCAR():
    """Test write_XDATCAR."""
    filename_vasprun = cwd / "vasprun.xml.tar.bz2"
    _tar = tarfile.open(filename_vasprun)
    for _, member in enumerate(_tar.getmembers()):
        vasprun = VasprunxmlExpat(_tar.extractfile(member))
        vasprun.parse()
        break

    with tempfile.TemporaryFile() as fp:
        write_XDATCAR(vasprunxml_expat=vasprun, fileptr=fp)
        fp.seek(0)
        lattice, positions = read_XDATCAR(fileptr=fp)

    np.testing.assert_allclose(lattice, np.eye(3) * 11.38060295, atol=1e-8)
    np.testing.assert_allclose(positions[0, 0], [0.00087869, 0, 0], atol=1e-8)
    np.testing.assert_allclose(positions[-1, -1], [0.5, 0.5, 0.75], atol=1e-8)



"""NaCl example to obtain dynamical matrix."""

import numpy as np
import yaml

import phonopy

phonon = phonopy.load(
    supercell_matrix=[[2, 0, 0], [0, 2, 0], [0, 0, 2]],
    primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
    unitcell_filename="POSCAR-unitcell",
    force_sets_filename="FORCE_SETS",
    born_filename="BORN",
)

q = [0.1, 0.1, 0.1]
dynmat = phonon.get_dynamical_matrix_at_q(q)
print(dynmat)
phonon.run_qpoints(q, with_dynamical_matrices=True)
print(phonon.get_qpoints_dict()["frequencies"][0])
phonon.write_yaml_qpoints_phonon()

data = yaml.load(open("qpoints.yaml"), Loader=yaml.FullLoader)
dynmat_from_yaml = []
dynmat_data = data["phonon"][0]["dynamical_matrix"]
for row in dynmat_data:
    vals = np.reshape(row, (-1, 2))
    dynmat_from_yaml.append(vals[:, 0] + vals[:, 1] * 1j)
dynmat_from_yaml = np.array(dynmat_from_yaml)
print(dynmat_from_yaml)
(
    eigvals,
    eigvecs,
) = np.linalg.eigh(dynmat)
frequencies = np.sqrt(np.abs(eigvals.real)) * np.sign(eigvals.real)
conversion_factor_to_THz = 15.633302
print(frequencies * conversion_factor_to_THz)



# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# PDOS plot (pdosplot)
#
# Usage:
#   pdosplot -i "1 2, 4 5" -o "pdos.pdf"
#
# The axis resolved PDOS is summed up with the successive
# indices separated by ",". In this example, indices 1 and
# 2, 3 and 4 are summed respectively, and then they are
# ploted respectively.
#
# The indices are defined like:
# 1 2 3 : X Y Z of the 1st atom,
# 4 5 6 : X Y Z of the 2nd atom,
# ...

import matplotlib.pyplot as plt
import numpy as np


def get_options():
    """Parse command-line options."""
    import argparse

    parser = argparse.ArgumentParser(description="Phonopy pdosplot command-line-tool")
    parser.set_defaults(
        output_filename=None,
        factor=1.0,
        legend_labels=None,
        xlabel=None,
        ylabel=None,
        show_legend=False,
        pdos_indices=None,
        ymax=None,
        ymin=None,
        title=None,
        f_max=None,
        f_min=None,
    )
    parser.add_argument(
        "--factor", dest="factor", type=float, help="Factor is multiplied with DOS."
    )
    parser.add_argument(
        "-l", "--legend", dest="show_legend", action="store_true", help="Show legend"
    )
    parser.add_argument(
        "--legend_labels", dest="legend_labels", help="Set legend labels"
    )
    parser.add_argument("--xlabel", dest="xlabel", help="Set x label")
    parser.add_argument("--ylabel", dest="ylabel", help="Set y label")
    parser.add_argument(
        "-i", "--indices", dest="pdos_indices", help="Indices like 1 2, 3 4 5 6..."
    )
    parser.add_argument(
        "-o", "--output", dest="output_filename", help="Output filename"
    )
    parser.add_argument("-t", "--title", dest="title", help="Title of plot")
    parser.add_argument(
        "--ymax", dest="ymax", type=float, help="Maximum value of y axis"
    )
    parser.add_argument(
        "--ymin", dest="ymin", type=float, help="Minimum value of y axis"
    )
    parser.add_argument(
        "--fmax", dest="f_max", type=float, help="Maximum frequency plotted"
    )
    parser.add_argument(
        "--fmin", dest="f_min", type=float, help="Minimum frequency plotted"
    )
    parser.add_argument(
        "filename", nargs="*", help="Filename of phonon DOS result (partial_dos.dat)"
    )
    args = parser.parse_args()
    return args


def run():
    """Run phonopy-pdosplot."""
    args = get_options()

    frequencies = []
    dos = []
    filename = "partial_dos.dat"
    if len(args.filename) > 0:
        filename = args.filename[0]
    for line in open(filename):
        if line.strip().split()[0] == "#" or line.strip().split() == "":
            continue

        tmp_array = [float(x) for x in line.split()]
        frequencies.append(tmp_array.pop(0))
        dos.append(tmp_array)

    frequencies = np.array(frequencies)
    dos = np.array(dos).transpose()

    # Extract indices
    indices = []
    if args.pdos_indices is None:
        indices.append(range(1, dos.shape[0] + 1))
    else:
        for v in args.pdos_indices.split(","):
            indices.append([int(x) for x in v.split()])

    # Set plot range in frequency axis
    if args.f_max is None:
        max_freq = max(frequencies)
    else:
        max_freq = args.f_max
    if args.f_min is None:
        min_freq = min(frequencies)
    else:
        min_freq = args.f_min

    min_i = 0
    max_i = len(frequencies)

    for i, f in enumerate(frequencies):
        if f > max_freq + (frequencies[1] - frequencies[0]) / 10:
            max_i = i + 1
            break

    for i, f in enumerate(frequencies):
        if f > min_freq - (frequencies[1] - frequencies[0]) / 10:
            min_i = i
            break

    # Plot
    fig, ax = plt.subplots()
    ax.xaxis.set_ticks_position("both")
    ax.yaxis.set_ticks_position("both")
    ax.xaxis.set_tick_params(which="both", direction="in")
    ax.yaxis.set_tick_params(which="both", direction="in")

    plots = []
    for nums in indices:
        pdos = np.zeros(frequencies.shape[0], dtype="double")
        for v in nums:
            pdos += dos[v - 1]
        (curve,) = plt.plot(frequencies[min_i:max_i], pdos[min_i:max_i] * args.factor)
        plots.append(curve)

    # plt.grid(True)

    ax.set_ylim((0, None))
    plt.xlim(min_freq, max_freq)

    if (args.ymin is not None) and (args.ymax is not None):
        plt.ylim(args.ymin, args.ymax)
    elif (args.ymin is None) and (args.ymax is not None):
        plt.ylim(ymax=args.ymax)
    elif (args.ymin is not None) and (args.ymax is None):
        plt.ylim(ymin=args.ymin)

    if args.xlabel is None:
        plt.xlabel("Frequency")
    else:
        plt.xlabel(args.xlabel)
    if args.ylabel is None:
        plt.ylabel("Partial density of states")
    else:
        plt.ylabel(args.ylabel)

    if args.show_legend:
        if args.legend_labels is not None:
            if len(args.legend_labels.split()) == len(plots):
                labels = args.legend_labels.split()
            else:
                print("Number of labels is not same as number of plots.")
                labels = indices
        else:
            labels = indices
        plt.legend(plots, labels, loc="upper left")

    if args.title is not None:
        plt.title(args.title)

    if args.output_filename is not None:
        plt.rcParams["pdf.fonttype"] = 42
        plt.rcParams["font.family"] = "serif"
        plt.savefig(args.output_filename)
    else:
        plt.show()



"""Example to calculate mode Grueneisen parameters."""

from __future__ import annotations

import numpy as np

from phonopy import Phonopy, PhonopyGruneisen
from phonopy.file_IO import parse_FORCE_SETS
from phonopy.interface.vasp import read_vasp


def _append_band(bands: list[list], q_start, q_end):
    band = []
    for i in range(51):
        points = np.array(q_start) + (np.array(q_end) - np.array(q_start)) / 50 * i
        band.append(points.tolist())
    bands.append(band)


phonons = {}
for vol in ("orig", "plus", "minus"):
    unitcell = read_vasp("%s/POSCAR-unitcell" % vol)
    phonon = Phonopy(
        unitcell,
        [[2, 0, 0], [0, 2, 0], [0, 0, 2]],
        primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
    )
    force_sets = parse_FORCE_SETS(filename="%s/FORCE_SETS" % vol)
    phonon.set_displacement_dataset(force_sets)
    phonon.produce_force_constants()
    phonons[vol] = phonon

gruneisen = PhonopyGruneisen(phonons["orig"], phonons["plus"], phonons["minus"])

gruneisen.set_mesh([2, 2, 2])
q_points, _, frequencies, _, gammas = gruneisen.get_mesh()
for q, freq, g in zip(q_points, frequencies, gammas):
    print(
        ("%5.2f %5.2f %5.2f " + (" %7.3f" * len(freq)))
        % ((q[0], q[1], q[2]) + tuple(freq))
    )
    print(((" " * 18) + (" %7.3f" * len(g))) % tuple(g))

bands: list[list] = []
_append_band(bands, [0.5, 0.5, 0.0], [0.0, 0.0, 0.0])
_append_band(bands, [0.0, 0.0, 0.0], [0.5, 0.5, 0.5])
gruneisen.set_band_structure(bands)

q_points, distances, frequencies, _, gammas = gruneisen.get_band_structure()
for q_path, d_path, freq_path, g_path in zip(q_points, distances, frequencies, gammas):
    for q, d, freq, g in zip(q_path, d_path, freq_path, g_path):
        print(
            ("%10.5f  %5.2f %5.2f %5.2f " + (" %7.3f" * len(freq)))
            % ((d, q[0], q[1], q[2]) + tuple(freq))
        )
        print(((" " * 30) + (" %7.3f" * len(g))) % tuple(g))



"""Example by corundum Al2O3."""

import numpy as np

import phonopy

phonon = phonopy.load(
    unitcell_filename="POSCAR-unitcell", supercell_matrix=[2, 2, 1], log_level=1
)
print("Space group: %s" % phonon.symmetry.get_international_table())

# Example to obtain dynamical matrix
dmat = phonon.get_dynamical_matrix_at_q([0, 0, 0])
print(dmat)

# Example of band structure calculation
bands = []
q_start = np.array([1.0 / 3, 1.0 / 3, 0])
q_end = np.array([0, 0, 0])
band = []
for i in range(51):
    band.append(q_start + (q_end - q_start) / 50 * i)
bands.append(band)

q_start = np.array([0, 0, 0])
q_end = np.array([1.0 / 3, 1.0 / 3, 1.0 / 2])
band = []
for i in range(51):
    band.append(q_start + (q_end - q_start) / 50 * i)
bands.append(band)

print("\nPhonon dispersion:")
phonon.run_band_structure(bands, with_eigenvectors=True, labels=["X", r"$\Gamma$", "L"])
band_plot = phonon.plot_band_structure()
band_plot.show()

bs = phonon.get_band_structure_dict()
distances = bs["distances"]
frequencies = bs["frequencies"]
qpoints = bs["qpoints"]

for qs_at_segments, dists_at_segments, freqs_at_segments in zip(
    qpoints, distances, frequencies
):
    for q, d, f in zip(qs_at_segments, dists_at_segments, freqs_at_segments):
        print("# %f %f %f" % tuple(q))
        print(("%s " + "%f " * len(f)) % ((d,) + tuple(f)))



"""Phonopy command line argument parser."""

# Copyright (C) 2016 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys


def fix_deprecated_option_names(argv):
    """Replace underscore in command option name by hyphen."""
    deprecated = []
    for i, v in enumerate(argv[1:]):
        if v[0] == "-":
            tag = v.split("=")[0]
            if "_" in tag:
                correct_tag = tag.replace("_", "-")
                deprecated.append(tag)
                argv[i + 1] = v.replace(tag, correct_tag)

    return deprecated


def show_deprecated_option_warnings(deprecated):
    """Show warning when underscore is included in command option name."""
    lines = [
        "Option names with underscores are deprecated, by which",
        "the underscores are replaced by dashes. Therefore",
    ]
    for tag in deprecated:
        lines.append("'%s' has to be written as '%s'." % (tag, tag.replace("_", "-")))
    maxlen = max([len(line) for line in lines])
    print("*" * maxlen)
    print("\n".join(lines))
    print("*" * maxlen)
    print("")


def get_parser(fc_symmetry=False, is_nac=False, load_phonopy_yaml=False):
    """Return ArgumentParser instance."""
    deprecated = fix_deprecated_option_names(sys.argv)
    import argparse

    from phonopy.interface.calculator import (
        add_arguments_of_calculators,
        calculator_info,
    )

    try:
        parser = argparse.ArgumentParser(
            description="Phonopy command-line-tool", allow_abbrev=False
        )  # allow_abbrev requires >= python 3.5
    except TypeError:
        parser = argparse.ArgumentParser(description="Phonopy command-line-tool")

    add_arguments_of_calculators(parser, calculator_info)

    parser.add_argument(
        "--alm",
        dest="use_alm",
        action="store_true",
        default=None,
        help="Use ALM for generating force constants",
    )
    parser.add_argument(
        "--amplitude",
        "--amin",
        dest="displacement_distance",
        type=float,
        default=None,
        help=(
            "Distance of displacements and also minimum distance of displacements "
            "in random displacements"
        ),
    )
    parser.add_argument(
        "--amax",
        dest="displacement_distance_max",
        type=float,
        default=None,
        help="Minimum distance of displacements in random displacements",
    )
    parser.add_argument(
        "--anime", nargs="+", dest="anime", default=None, help="Same as ANIME tag"
    )
    parser.add_argument(
        "--band",
        nargs="+",
        dest="band_paths",
        default=None,
        help="Same behavior as BAND tag",
    )
    parser.add_argument(
        "--band-connection",
        dest="is_band_connection",
        action="store_true",
        default=None,
        help="Treat band crossings",
    )
    parser.add_argument(
        "--band-const-interval",
        dest="is_band_const_interval",
        action="store_true",
        default=None,
        help="Band paths are sampled with similar interval.",
    )
    parser.add_argument(
        "--band-labels",
        nargs="+",
        dest="band_labels",
        default=None,
        help="Show labels at band segments",
    )
    parser.add_argument(
        "--band-format",
        dest="band_format",
        default=None,
        help="Band structure output file-format",
    )
    parser.add_argument(
        "--band-points",
        dest="band_points",
        type=int,
        default=None,
        help=(
            "Number of points calculated on a band segment in the band structure mode"
        ),
    )
    parser.add_argument(
        "--bi",
        "--band-indices",
        nargs="+",
        dest="band_indices",
        default=None,
        help=("Band indices to be included to calcualte thermal properties"),
    )
    if not load_phonopy_yaml:
        parser.add_argument(
            "-c",
            "--cell",
            dest="cell_filename",
            metavar="FILE",
            default=None,
            help="Read unit cell",
        )
    if load_phonopy_yaml:
        parser.add_argument(
            "--config",
            dest="conf_filename",
            metavar="FILE",
            default=None,
            help="Phonopy configuration file",
        )
    parser.add_argument(
        "--classical",
        dest="classical",
        action="store_true",
        default=False,
        help=("Compute thermodynamic properties using classical statistics."),
    )
    parser.add_argument(
        "--cutoff-freq",
        "--cutoff-frequency",
        dest="cutoff_frequency",
        type=float,
        default=None,
        help=("Thermal properties are not calculated below this cutoff frequency."),
    )
    parser.add_argument(
        "--cutoff-radius",
        dest="cutoff_radius",
        type=float,
        default=None,
        help="Out of cutoff radius, force constants are set zero.",
    )
    parser.add_argument(
        "-d",
        "--displacement",
        dest="is_displacement",
        action="store_true",
        default=None,
        help="Create supercells with displacements",
    )
    if not load_phonopy_yaml:
        parser.add_argument(
            "--dim",
            nargs="+",
            dest="supercell_dimension",
            default=None,
            help="Same behavior as DIM tag",
        )
    parser.add_argument(
        "--dm-decimals",
        dest="dynamical_matrix_decimals",
        default=None,
        type=int,
        help="Decimals of values of decimals",
    )
    parser.add_argument(
        "--dos",
        dest="is_dos_mode",
        action="store_true",
        default=None,
        help="Calculate (P)DOS",
    )
    parser.add_argument(
        "--eigvecs",
        "--eigenvectors",
        dest="is_eigenvectors",
        action="store_true",
        default=None,
        help="Output eigenvectors",
    )
    if load_phonopy_yaml:
        parser.add_argument(
            "--exclude-born",
            "--exclude-nac-params",
            dest="include_nac_params",
            action="store_false",
            default=None,
            help=(
                "Exclude born effective charge and dielectric tensor in phonopy.yaml"
            ),
        )
    parser.add_argument(
        "-f",
        "--force-sets",
        nargs="+",
        dest="create_force_sets",
        default=None,
        help="Create FORCE_SETS",
    )
    parser.add_argument(
        "--factor",
        dest="frequency_conversion_factor",
        type=float,
        default=None,
        help="Frequency unit conversion factor",
    )
    parser.add_argument(
        "--fc",
        "--force-constants",
        metavar="FILE",
        dest="create_force_constants",
        default=None,
        help=(
            "Create FORCE_CONSTANTS from vaspurn.xml. "
            "vasprun.xml has to be passed as argument."
        ),
    )
    parser.add_argument(
        "--fc-calc",
        "--fc-calculator",
        dest="fc_calculator",
        default=None,
        help=("Force constants calculator"),
    )
    parser.add_argument(
        "--fc-calc-opt",
        "--fc-calculator-options",
        dest="fc_calculator_options",
        default=None,
        help=(
            "Options for force constants calculator as comma separated "
            "string with the style of key = values"
        ),
    )
    parser.add_argument(
        "--fc-decimals",
        dest="force_constants_decimals",
        type=int,
        default=None,
        help="Decimals of values of force constants",
    )
    parser.add_argument(
        "--fc-format",
        dest="fc_format",
        default=None,
        help="Force constants input/output file-format",
    )
    parser.add_argument(
        "--fc-spg-symmetry",
        dest="fc_spg_symmetry",
        action="store_true",
        default=None,
        help="Enforce space group symmetry to force constants",
    )
    if not fc_symmetry:
        parser.add_argument(
            "--fc-symmetry",
            "--sym-fc",
            dest="fc_symmetry",
            action="store_true",
            default=None,
            help="Symmetrize force constants",
        )
    parser.add_argument(
        "--fits-debye-model",
        dest="fits_debye_model",
        action="store_true",
        default=None,
        help="Fits total DOS to a Debye model",
    )
    # parser.add_argument(
    #     "--freq-scale",
    #     dest="frequency_scale_factor",
    #     type=float,
    #     default=None,
    #     help=(
    #         "Squared scale factor multiplied as fc2 * factor^2. Therefore "
    #         "frequency is changed but the contribution from NAC is not "
    #         "changed."
    #     ),
    # )
    parser.add_argument(
        "--full-fc",
        dest="is_full_fc",
        action="store_true",
        default=None,
        help="Calculate full supercell force constants matrix",
    )
    parser.add_argument(
        "--fz",
        "--force-sets-zero",
        nargs="+",
        dest="create_force_sets_zero",
        default=None,
        help=(
            "Create FORCE_SETS. disp.yaml in the current directory and "
            "vapsrun.xml's for VASP or case.scf(m) for Wien2k as arguments "
            "are required. The first argument is that of the perfect "
            "supercell to subtract residual forces"
        ),
    )
    parser.add_argument(
        "--fmax",
        dest="fmax",
        type=float,
        default=None,
        help="Maximum frequency used for DOS or moment calculation",
    )
    parser.add_argument(
        "--fmin",
        dest="fmin",
        type=float,
        default=None,
        help="Minimum frequency used for DOS or moment calculation",
    )
    parser.add_argument(
        "--fpitch",
        dest="fpitch",
        type=float,
        help="Frequency pitch used for DOS or moment calculation",
    )
    parser.add_argument(
        "--gc",
        "--gamma-center",
        dest="is_gamma_center",
        action="store_true",
        default=None,
        help="Set mesh as Gamma center",
    )
    parser.add_argument(
        "--gv",
        "--group-velocity",
        dest="is_group_velocity",
        action="store_true",
        default=None,
        help="Calculate group velocities at q-points",
    )
    parser.add_argument(
        "--gv-delta-q",
        dest="gv_delta_q",
        type=float,
        default=None,
        help="Delta-q distance used for group velocity calculation",
    )
    parser.add_argument(
        "--hdf5",
        dest="is_hdf5",
        action="store_true",
        default=None,
        help="Use hdf5 for force constants",
    )
    parser.add_argument(
        "--hdf5-compression",
        dest="hdf5_compression",
        default=None,
        help="hdf5 compression filter (default: gzip)",
    )
    parser.add_argument(
        "--irreps",
        "--irreps-qpoint",
        nargs="+",
        dest="irreps_qpoint",
        default=None,
        help="A q-point where characters of irreps are calculated",
    )
    parser.add_argument(
        "--include-fc",
        dest="include_fc",
        action="store_true",
        default=None,
        help="Include force constants in phonopy.yaml",
    )
    parser.add_argument(
        "--include-fs",
        dest="include_fs",
        action="store_true",
        default=None,
        help="Include force sets in phonopy.yaml",
    )
    parser.add_argument(
        "--include-disp",
        dest="include_disp",
        action="store_true",
        default=None,
        help="Include displacements in phonopy.yaml",
    )
    parser.add_argument(
        "--include-all",
        dest="include_all",
        action="store_true",
        default=None,
        help="Include all output file data in phonopy.yaml",
    )
    parser.add_argument(
        "--legend",
        dest="is_legend",
        action="store_true",
        default=None,
        help="Legend of plots is shown in thermal displacements",
    )
    parser.add_argument(
        "--legacy-plot",
        dest="is_legacy_plot",
        action="store_true",
        default=None,
        help="Legacy style band structure pl",
    )
    parser.add_argument(
        "--lcg",
        "--little-cogroup",
        dest="is_little_cogroup",
        action="store_true",
        default=None,
        help=(
            "Show irreps of little co-group (or point-group of "
            "wave vector q) instead of little group"
        ),
    )
    parser.add_argument(
        "--loglevel", dest="loglevel", type=int, default=None, help="Log level"
    )
    parser.add_argument(
        "--mass", nargs="+", dest="masses", default=None, help="Same as MASS tag"
    )
    parser.add_argument(
        "--magmom", nargs="+", dest="magmoms", default=None, help="Same as MAGMOM tag"
    )
    parser.add_argument(
        "--mesh-format",
        dest="mesh_format",
        default=None,
        help="Mesh output file-format",
    )
    parser.add_argument(
        "--modulation",
        nargs="+",
        dest="modulation",
        default=None,
        help="Same as MODULATION tag",
    )
    parser.add_argument(
        "--mp",
        "--mesh",
        nargs="+",
        dest="mesh_numbers",
        default=None,
        help="Same behavior as MP tag",
    )
    parser.add_argument(
        "--mlp-params",
        dest="mlp_params",
        default=None,
        help=(
            "Parameters for machine learning potentials as comma separated "
            "string with the style of key = values"
        ),
    )
    parser.add_argument(
        "--moment",
        dest="is_moment",
        action="store_true",
        default=None,
        help="Calculate moment of phonon states distribution",
    )
    parser.add_argument(
        "--moment-order",
        dest="moment_order",
        default=None,
        type=int,
        help="Order of moment of phonon states distribution",
    )
    if not is_nac:
        parser.add_argument(
            "--nac",
            dest="is_nac",
            action="store_true",
            default=None,
            help="Non-analytical term correction",
        )
    parser.add_argument(
        "--nac-method",
        dest="nac_method",
        default=None,
        help="Non-analytical term correction method: Gonze (default) or Wang",
    )
    if fc_symmetry:
        parser.add_argument(
            "--no-fc-symmetry",
            "--no-sym-fc",
            dest="fc_symmetry",
            action="store_false",
            default=None,
            help="Do not symmetrize force constants",
        )
    parser.add_argument(
        "--nodiag",
        dest="is_nodiag",
        action="store_true",
        default=None,
        help="Set displacements parallel to axes",
    )
    parser.add_argument(
        "--nomeshsym",
        dest="is_nomeshsym",
        action="store_true",
        default=None,
        help="Symmetry is not imposed for mesh sampling.",
    )
    if is_nac:
        parser.add_argument(
            "--nonac",
            dest="is_nac",
            action="store_false",
            default=None,
            help="Non-analytical term correction",
        )
    parser.add_argument(
        "--nosym",
        dest="is_nosym",
        action="store_true",
        default=None,
        help="Symmetry is not imposed.",
    )
    parser.add_argument(
        "--nowritemesh",
        dest="write_mesh",
        action="store_false",
        default=None,
        help="Do not write mesh.yaml or mesh.hdf5",
    )
    parser.add_argument(
        "-p",
        "--plot",
        dest="is_graph_plot",
        action="store_true",
        default=None,
        help="Plot data",
    )
    parser.add_argument(
        "--pa",
        "--primitive-axis",
        "--primitive-axes",
        nargs="+",
        dest="primitive_axes",
        default=None,
        help="Same as PRIMITIVE_AXES tag",
    )
    parser.add_argument(
        "--pd",
        "--projection-direction",
        nargs="+",
        dest="projection_direction",
        default=None,
        help="Same as PROJECTION_DIRECTION tag",
    )
    parser.add_argument(
        "--pdos", nargs="+", dest="pdos", default=None, help="Same as PDOS tag"
    )
    parser.add_argument(
        "--pm",
        dest="is_plusminus_displacements",
        action="store_true",
        default=None,
        help="Set plus minus displacements",
    )
    parser.add_argument(
        "--pr",
        "--pretend-real",
        dest="pretend_real",
        action="store_true",
        default=None,
        help=(
            "Use imaginary frequency as real for thermal property "
            "calculation. For a testing purpose only, when a small "
            "amount of imaginary branches obtained."
        ),
    )
    parser.add_argument(
        "--pt",
        "--projected-thermal-property",
        dest="is_projected_thermal_properties",
        action="store_true",
        default=None,
        help="Output projected thermal properties",
    )
    parser.add_argument(
        "--pypolymlp",
        dest="use_pypolymlp",
        action="store_true",
        default=None,
        help="Use pypolymlp for generating force constants",
    )
    parser.add_argument(
        "--qpoints",
        nargs="+",
        dest="qpoints",
        default=None,
        help="Calculate at specified q-points",
    )
    parser.add_argument(
        "--qpoints-format",
        dest="qpoints_format",
        default=None,
        help="Q-points output file-format",
    )
    parser.add_argument(
        "--q-direction",
        nargs="+",
        dest="nac_q_direction",
        default=None,
        help=(
            "Direction of q-vector perturbation used for NAC at "
            "q->0, and group velocity for degenerate phonon "
            "mode in q-points mode"
        ),
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        default=None,
        help="Print out smallest information",
    )
    parser.add_argument(
        "--random-seed",
        dest="random_seed",
        type=int,
        default=None,
        help="Random seed by a 32 bit unsigned integer",
    )
    parser.add_argument(
        "--rd",
        "--random-displacements",
        dest="random_displacements",
        default=None,
        help='Number of supercells with random displacements or "auto".',
    )
    parser.add_argument(
        "--rd-temperature",
        dest="rd_temperature",
        type=float,
        default=None,
        metavar="TEMPERATURE",
        help="A temperature used to generate random displacements.",
    )
    parser.add_argument(
        "--temperature",
        dest="temperature",
        type=float,
        default=None,
        metavar="TEMPERATURE",
        help="(Deprecated) A temperature used to generate random displacements.",
    )
    parser.add_argument(
        "--readfc",
        dest="read_force_constants",
        action="store_true",
        default=None,
        help="Read FORCE_CONSTANTS",
    )
    parser.add_argument(
        "--readfc-format",
        dest="readfc_format",
        default=None,
        help="Force constants input file-format",
    )
    parser.add_argument(
        "--read-qpoints",
        dest="read_qpoints",
        action="store_true",
        default=None,
        help="Read QPOITNS",
    )
    parser.add_argument(
        "-s",
        "--save",
        dest="is_graph_save",
        action="store_true",
        default=None,
        help="Save plot data in pdf",
    )
    parser.add_argument(
        "--sp",
        "--save-params",
        dest="save_params",
        action="store_true",
        default=None,
        help="Save parameters that can run phonopy in phonopy_params.yaml.",
    )
    parser.add_argument(
        "--show-irreps",
        dest="show_irreps",
        action="store_true",
        default=None,
        help="Show IR-Reps along with characters",
    )
    parser.add_argument(
        "--sigma", dest="sigma", default=None, help="Smearing width for DOS"
    )
    parser.add_argument(
        "--sscha",
        dest="sscha_iterations",
        type=int,
        default=None,
        help="Number of iterations in SSCHA calculation",
    )
    if not fc_symmetry:
        parser.add_argument(
            "--symfc",
            dest="use_symfc",
            action="store_true",
            default=None,
            help="Use symfc for generating force constants",
        )
    parser.add_argument(
        "--symmetry",
        dest="is_check_symmetry",
        action="store_true",
        default=None,
        help="Check crystal symmetry",
    )
    parser.add_argument(
        "-t",
        "--thermal-property",
        dest="is_thermal_properties",
        action="store_true",
        default=None,
        help="Output thermal properties",
    )
    parser.add_argument(
        "--td",
        "--thermal-displacements",
        dest="is_thermal_displacements",
        action="store_true",
        default=None,
        help="Output thermal displacements",
    )
    parser.add_argument(
        "--tdm",
        "--thermal-displacement-matrix",
        dest="is_thermal_displacement_matrices",
        action="store_true",
        default=None,
        help="Output thermal displacement matrices",
    )
    parser.add_argument(
        "--tdm-cif",
        "--thermal-displacement-matrix-cif",
        metavar="TEMPERATURE",
        dest="thermal_displacement_matrices_cif",
        type=float,
        default=None,
        help="Write cif with aniso_U for which temperature is specified",
    )
    parser.add_argument(
        "--tmax",
        dest="tmax",
        type=float,
        default=None,
        help="Maximum calculated temperature",
    )
    parser.add_argument(
        "--tmin",
        dest="tmin",
        type=float,
        default=None,
        help="Minimum calculated temperature",
    )
    parser.add_argument(
        "--tolerance",
        dest="symmetry_tolerance",
        type=float,
        default=None,
        help="Symmetry tolerance to search",
    )
    parser.add_argument(
        "--trigonal",
        dest="is_trigonal_displacements",
        action="store_true",
        default=None,
        help="Set displacements of all trigonal axes ",
    )
    parser.add_argument(
        "--tstep",
        dest="tstep",
        type=float,
        default=None,
        help="Calculated temperature step",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=None,
        help="Detailed information is shown.",
    )
    parser.add_argument(
        "--wien2k-p1",
        dest="is_wien2k_p1",
        action="store_true",
        default=None,
        help="Assume Wien2k structs with displacements are P1",
    )
    parser.add_argument(
        "--writefc",
        dest="write_force_constants",
        action="store_true",
        default=None,
        help="Write FORCE_CONSTANTS",
    )
    parser.add_argument(
        "--writefc-format",
        dest="writefc_format",
        default=None,
        help="Force constants output file-format",
    )
    parser.add_argument(
        "--writedm",
        dest="write_dynamical_matrices",
        action="store_true",
        default=None,
        help=(
            "Write dynamical matrices. This has to be used "
            "with QPOINTS setting (or --qpoints)"
        ),
    )
    parser.add_argument(
        "--xyz-projection",
        dest="xyz_projection",
        action="store_true",
        default=None,
        help="Project PDOS x, y, z directions in Cartesian coordinates",
    )
    if load_phonopy_yaml:
        parser.add_argument("filename", nargs="*", help="phonopy.yaml like file")
    else:
        parser.add_argument(
            "filename",
            nargs="*",
            help=(
                "Phonopy configure file. However if the file is recognized as "
                "phonopy.yaml like file, this file is read as phonopy.yaml like file."
            ),
        )

    return parser, deprecated



"""LAMMPS calculator interface."""

# Copyright (C) 2023 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import io
import os
import re
from typing import Union

import numpy as np

from phonopy.interface.vasp import check_forces, get_drift_forces
from phonopy.structure.atoms import PhonopyAtoms, atom_data
from phonopy.structure.cells import get_cell_matrix_from_lattice


def write_supercells_with_displacements(
    supercell,
    cells_with_displacements,
    ids,
    pre_filename="supercell",
    width=3,
):
    """Write supercells with displacements to files."""
    write_lammps(pre_filename, supercell)
    for i, cell in zip(ids, cells_with_displacements):
        filename = f"{pre_filename}-{i:0{width}d}"
        write_lammps(filename, cell)


def write_lammps(filename, cell):
    """Write LAMMPS structure to file."""
    with open(filename, "w") as w:
        w.write("\n".join(LammpsStructureDumper(cell).get_lines()))


def read_lammps(cell_filename) -> PhonopyAtoms:
    """Read LAMMPS structure file and return cell."""
    return LammpsStructureLoader().load(cell_filename).cell


def parse_set_of_forces(num_atoms: int, forces_filenames: list, verbose: bool = True):
    """Parse forces from output files."""
    is_parsed = True
    force_sets = []

    for i, filename in enumerate(forces_filenames):
        if verbose:
            print(f"{i + 1}. ", end="")
        forces = LammpsForcesLoader().load(filename).forces
        if check_forces(forces, num_atoms, filename, verbose=verbose):
            drift_force = get_drift_forces(forces, filename=filename, verbose=verbose)
            force_sets.append(np.array(forces) - drift_force)
        else:
            is_parsed = False

    if is_parsed:
        return force_sets
    else:
        return []


def rotate_lammps_forces(force_sets: list, lattice: np.ndarray, verbose: bool = True):
    """Rotate forces of LAMMPS output.

    Parameters
    ----------
    force_sets : list
        Sets of supercell forces obtained by ``parse_set_of_forces``.
    lattice : ndarray
        Basis vectors in row vectors.
    verbose : bool
        Verbosity.

    """
    lat = lattice
    rot_lat = get_cell_matrix_from_lattice(lat)
    r = np.dot(np.linalg.inv(lat), rot_lat).T
    for forces in force_sets:
        # (Rinv.f)^T = f^T.Rinv^T = f^T.R
        forces[:] = np.dot(forces, r)
    if verbose:
        print("Forces parsed from LAMMPS output were rotated by F=R.F(lammps) with R:")
        for v in r.T:
            print(f"  {v[0]:7.5f} {v[1]:7.5f} {v[2]:7.5f}")


class LammpsStructureDumper:
    """Class to create LAMMPS input structure file.

    LAMMPS requires the basis vectors to be defined in the following way

    a = (a_x 0 0)
    b = (b_x b_y 0)
    c = (c_x c_y c_z)

    So the ``cell.cell`` has to be rotated to match this style. The coordinates
    of atoms have to be given in Cartesian coordinates with respect to this
    set of basis vectors.

    """

    def __init__(self, cell: PhonopyAtoms):
        """Init method."""
        lattice = get_cell_matrix_from_lattice(cell.cell)
        lmps_cell = PhonopyAtoms(
            cell=lattice, scaled_positions=cell.scaled_positions, symbols=cell.symbols
        )
        self._run(lmps_cell)

    def get_lines(self) -> list[str]:
        """Return LAMMPS structure str lines."""
        return self._lines

    def _run(self, cell: PhonopyAtoms):
        unums, uids = np.unique(cell.numbers, return_index=True)
        usyms = [cell.symbols[i] for i in uids]
        num_map = {n: i for i, n in enumerate(unums)}

        lines = ["#", ""]
        lines.append(f"{len(cell.numbers)} atoms")
        lines.append(f"{len(np.unique(cell.numbers))} atom types")
        lines.append("")
        lines.append(f"0.0 {cell.cell[0, 0]} xlo xhi")
        lines.append(f"0.0 {cell.cell[1, 1]} ylo yhi")
        lines.append(f"0.0 {cell.cell[2, 2]} zlo zhi")
        lines.append("")
        lines.append(f"{cell.cell[1, 0]} {cell.cell[2, 0]} {cell.cell[2, 1]} xy xz yz")
        lines.append("")
        lines.append("")
        lines.append("Atom Type Labels")
        lines.append("")
        for i, symbol in enumerate(usyms):
            lines.append(f"{i + 1} {symbol}")
        lines.append("")
        lines.append("")
        lines.append("Atoms")
        lines.append("")
        for i, (position, number) in enumerate(zip(cell.positions, cell.numbers)):
            pos_str = f"{position[0]} {position[1]} {position[2]}"
            lines.append(f"{i + 1} {usyms[num_map[number]]} {pos_str}")
        self._lines = lines


class LammpsForcesLoader:
    """Class to load LAMMPS output file to get forces.

    Configuration by following three lines is assumed in the LAMMPS input script.

    dump phonopy all custom 1 forces.* id type x y z fx fy fz
    dump_modify phonopy format line "%d %d %15.8f %15.8f %15.8f %15.8f %15.8f %15.8f"
    run 0

    Then ``forces.0`` file is created as the result of LAMMPS calculation. Forces
    are parsed from lines after:

    ITEM: ATOMS id type x y z fx fy fz

    The values of 6-8th columns are the forces in Cartesian coordinates.

    """

    _hooks = {
        "forces": r"ITEM:\s+ATOMS\s+id\s+type",
        "num_atoms": r"ITEM:\s+NUMBER\s+OF\s+ATOMS",
    }

    def __init__(self):
        """Init method."""
        self._forces = None

    def load(self, fp: Union[str, bytes, os.PathLike, io.IOBase]):
        """Load and parse LAMMPS structure file.

        Parameters
        ----------
        fp : filename or stream
            filename or stream.

        """
        return _load(self, fp, return_lines=False)

    @property
    def forces(self) -> np.ndarray:
        """Return forces."""
        return self._forces

    def _parse(self, fp: io.IOBase, column_start=5, column_end=8):
        """Parse lines of LAMMPS output file."""
        num_atoms = -1
        for line in fp:
            regex = re.compile(self._hooks["num_atoms"])
            if regex.search(line):
                num_atoms = int(fp.readline().strip())
                break

        for line in fp:
            regex = re.compile(self._hooks["forces"])
            if regex.search(line):
                break

        forces = np.zeros((num_atoms, 3), dtype="double")
        indices_found = [False] * num_atoms
        for i, line in enumerate(fp):
            if i == num_atoms:
                break
            ary = line.split()
            atom_id = int(ary[0])
            indices_found[atom_id - 1] = True
            forces[atom_id - 1] = np.array(ary[column_start:column_end], dtype="double")

        assert all(indices_found)
        self._forces = forces


class LammpsStructureLoader:
    """Class to load LAMMPS input structure file.

    lmps = LammpsStructureLoader()
    lmps.parse(lines)
    cell: PhonopyAtoms = lmps.cell

    """

    _header_hooks = {
        "xlo_xhi": (re.compile(r"xlo\s+xhi"), "_set_xlo_xhi"),
        "ylo_yhi": (re.compile(r"ylo\s+yhi"), "_set_xlo_xhi"),
        "zlo_zhi": (re.compile(r"zlo\s+zhi"), "_set_xlo_xhi"),
        "xy_xz_yz": (re.compile(r"xy\s+xz\s+yz"), "_set_xlo_xhi"),
        "atoms": (re.compile("atoms"), "_set_number_of_atoms"),
        "atom_types": (re.compile(r"atom\s+types"), "_set_number_of_atoms"),
    }

    def __init__(self):
        """Init method."""
        self._header_tags = {}
        self._atom_type_labels = {}
        self._atom_ids = None
        self._atom_labels = None
        self._atom_positions = None
        self._cell = None

    @property
    def cell(self) -> PhonopyAtoms:
        """Return parsed cell."""
        return self._cell

    def load(self, fp: Union[str, bytes, os.PathLike, io.IOBase]):
        """Load and parse LAMMPS structure file.

        Parameters
        ----------
        fp : filename or stream
            filename or stream.

        """
        return _load(self, fp)

    def _parse(self, lines):
        """Parse LAMMPS structure file."""
        re_ATL = re.compile(r"Atom\s+Type\s+Labels")
        for line in lines:
            for key, (regex, method_name) in self._header_hooks.items():
                if regex.search(line):
                    getattr(self, method_name)(key, regex.sub("", line))
                    break

        i = 0
        while True:
            if i > len(lines):
                break
            line = lines[i]
            # Hook "Atom Type Labels"
            if re_ATL.search(line):
                i += self._parse_AtomTypeLabels(lines[(i + 1) :])
                continue
            # Hook "Atoms", this must be after "Atom Type Labels".
            if "Atoms" == line.split("#")[0].strip():
                self._parse_Atoms(lines[(i + 1) :])
                break
            i += 1

        lattice = np.zeros((3, 3), dtype="double", order="C")
        tag = self._header_tags
        lattice[0, 0] = tag["xlo_xhi"][1] - tag["xlo_xhi"][0]
        lattice[1, 1] = tag["ylo_yhi"][1] - tag["ylo_yhi"][0]
        lattice[2, 2] = tag["zlo_zhi"][1] - tag["zlo_zhi"][0]
        lattice[1, 0] = tag["xy_xz_yz"][0]  # xy
        lattice[2, 0] = tag["xy_xz_yz"][1]  # xz
        lattice[2, 1] = tag["xy_xz_yz"][2]  # yz

        if self._atom_type_labels:
            self._cell = PhonopyAtoms(
                cell=lattice, positions=self._atom_positions, symbols=self._atom_labels
            )
        else:
            symbols = [atom_data[n][1] for n in self._atom_labels]
            self._cell = PhonopyAtoms(
                cell=lattice, positions=self._atom_positions, symbols=symbols
            )

    def _parse_AtomTypeLabels(self, lines):
        num_types = 0
        for i, line in enumerate(lines):  # noqa: B007
            _line = line.split("#")[0].strip()
            if _line == "":
                continue
            ary = _line.split()
            self._atom_type_labels[ary[1]] = int(ary[0])
            num_types += 1
            if num_types == self._header_tags["atom_types"]:
                break
        assert num_types == self._header_tags["atom_types"]
        return i + 1

    def _parse_Atoms(self, lines):
        positions = np.zeros((self._header_tags["atoms"], 3), dtype="double", order="C")
        if self._atom_type_labels:
            lables = []
        else:
            lables = np.zeros(self._header_tags["atoms"], dtype="int64")
        ids = np.zeros(self._header_tags["atoms"], dtype="int64")
        num_atoms = 0
        for line in lines:
            _line = line.split("#")[0].strip()
            if _line == "":
                continue
            ary = _line.split()
            ids[num_atoms] = int(ary[0])

            # ary[1] can be chemical symbol (key) or id (val).
            # Supporting id is useful for "write_data" LAMMPS command.
            if self._atom_type_labels:
                for key, val in self._atom_type_labels.items():
                    if ary[1] in (key, f"{val}"):
                        lables.append(key)
                        break
            else:
                lables[num_atoms] = int(ary[1])
            positions[num_atoms] = [float(v) for v in ary[2:5]]
            num_atoms += 1
            if num_atoms == self._header_tags["atoms"]:
                break
        assert num_atoms == self._header_tags["atoms"], (
            f"{num_atoms} != {self._header_tags['atoms']}"
        )
        self._atom_ids = ids
        self._atom_labels = lables
        self._atom_positions = positions

    def _set_xlo_xhi(self, key, line):
        self._header_tags[key] = np.array(
            [float(v) for v in line.split("#")[0].split()]
        )

    def _set_number_of_atoms(self, key, line):
        self._header_tags[key] = int(line.split("#")[0])


def _load(self, fp: Union[str, bytes, os.PathLike, io.IOBase], return_lines=True):
    """Load and parse LAMMPS structure file.

    Parameters
    ----------
    fp : filename or stream
        filename or stream.

    """
    if isinstance(fp, io.IOBase):
        if return_lines:
            self._parse(fp.readlines())
        else:
            self._parse(fp)
    else:
        with open(fp) as f:
            if return_lines:
                self._parse(f.readlines())
            else:
                self._parse(f)
    return self



"""Example to read and write FORCE_CONSTANTS file."""

import phonopy
from phonopy.file_IO import parse_FORCE_CONSTANTS, write_FORCE_CONSTANTS

phonon = phonopy.load(
    supercell_matrix=[[2, 0, 0], [0, 2, 0], [0, 0, 2]],
    primitive_matrix=[[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]],
    unitcell_filename="POSCAR-unitcell",
    force_sets_filename="FORCE_SETS",
    born_filename="BORN",
)
write_FORCE_CONSTANTS(phonon.force_constants, filename="FORCE_CONSTANTS")

force_constants = parse_FORCE_CONSTANTS()
phonon.force_constants = force_constants
phonon.symmetrize_force_constants()
write_FORCE_CONSTANTS(phonon.force_constants, filename="FORCE_CONSTANTS_NEW")
